# Copyright (C) 2020 Jørgen S. Dokken and Chris N. Richardson
#
# This file is part of DOLFINx (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from mpi4py import MPI

import numpy as np
import pytest
from scipy.spatial.transform import Rotation

import ufl
from basix.ufl import element
from dolfinx import geometry
from dolfinx.geometry import compute_distance_gjk
from dolfinx.mesh import create_mesh


def distance_point_to_line_3D(P1, P2, point):
    """Distance from point to line."""
    return np.linalg.norm(np.cross(P2 - P1, P1 - point)) / np.linalg.norm(P2 - P1)


def distance_point_to_plane_3D(P1, P2, P3, point):
    """Distance from point to plane."""
    return np.abs(
        np.dot(np.cross(P2 - P1, P3 - P1) / np.linalg.norm(np.cross(P2 - P1, P3 - P1)), point - P2)
    )


@pytest.mark.parametrize("delta", [0.1, 1e-12, 0, -2])
@pytest.mark.parametrize("dtype", [np.float64, np.float32])
def test_line_point_distance(delta, dtype):
    line = np.array([[0.1, 0.2, 0.3], [0.5, 0.8, 0.7]], dtype=dtype)
    point_on_line = line[0] + 0.27 * (line[1] - line[0])
    normal = np.cross(line[0], line[1])
    point = point_on_line + delta * normal
    d = compute_distance_gjk(line, point)
    distance = np.linalg.norm(d)
    actual_distance = distance_point_to_line_3D(line[0], line[1], point)
    assert np.isclose(distance, actual_distance, atol=1e-8)


@pytest.mark.parametrize("delta", [0.1, 1e-12, 0])
@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_line_line_distance(delta, dtype):
    line = np.array([[-0.5, -0.7, -0.3], [1, 2, 3]], dtype=dtype)
    point_on_line = line[0] + 0.38 * (line[1] - line[0])
    normal = np.cross(line[0], line[1])
    point = point_on_line + delta * normal
    line_2 = np.array([point, [2, 5, 6]], dtype=dtype)
    distance = np.linalg.norm(compute_distance_gjk(line, line_2))
    actual_distance = distance_point_to_line_3D(line[0], line[1], line_2[0])
    assert np.isclose(distance, actual_distance, atol=1e-7)


@pytest.mark.parametrize("delta", [0.1 ** (3 * i) for i in range(6)])
@pytest.mark.parametrize("dtype", [np.float64])
def test_tri_distance(delta, dtype):
    tri_1 = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0]], dtype=dtype)
    tri_2 = np.array([[1, delta, 0], [3, 1.2, 0], [1, 1, 0]], dtype=dtype)
    P1 = tri_1[2]
    P2 = tri_1[1]
    point = tri_2[0]
    actual_distance = distance_point_to_line_3D(P1, P2, point)
    distance = np.linalg.norm(compute_distance_gjk(tri_1, tri_2))
    assert np.isclose(distance, actual_distance, atol=1e-6)


@pytest.mark.parametrize("delta", [0.1 * 0.1 ** (3 * i) for i in range(6)])
@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_quad_distance2d(delta, dtype):
    quad_1 = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0], [1, 1, 0]], dtype=dtype)
    quad_2 = np.array([[0, 1 + delta, 0], [2, 2, 0], [2, 4, 0], [4, 4, 0]], dtype=dtype)
    P1 = quad_1[2]
    P2 = quad_1[3]
    point = quad_2[0]
    actual_distance = distance_point_to_line_3D(P1, P2, point)
    distance = np.linalg.norm(compute_distance_gjk(quad_1, quad_2))
    assert np.isclose(distance, actual_distance, atol=1e-8)


@pytest.mark.parametrize("delta", [1 * 0.5 ** (3 * i) for i in range(7)])
def test_tetra_distance_3d(delta):
    tetra_1 = np.array([[0, 0, 0.2], [1, 0, 0.1], [0, 1, 0.3], [0, 0, 1]], dtype=np.float64)
    tetra_2 = np.array([[0, 0, -3], [1, 0, -3], [0, 1, -3], [0.5, 0.3, -delta]], dtype=np.float64)
    actual_distance = distance_point_to_plane_3D(tetra_1[0], tetra_1[1], tetra_1[2], tetra_2[3])
    distance = np.linalg.norm(compute_distance_gjk(tetra_1, tetra_2))
    assert np.isclose(distance, actual_distance, atol=1e-15)


@pytest.mark.parametrize("delta", [(-1) ** i * np.sqrt(2) * 0.1 ** (3 * i) for i in range(6)])
def test_tetra_collision_3d(delta):
    tetra_1 = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=np.float64)
    tetra_2 = np.array([[0, 0, -3], [1, 0, -3], [0, 1, -3], [0.5, 0.3, -delta]], dtype=np.float64)
    actual_distance = distance_point_to_plane_3D(tetra_1[0], tetra_1[1], tetra_1[2], tetra_2[3])
    distance = np.linalg.norm(compute_distance_gjk(tetra_1, tetra_2))
    if delta < 0:
        assert np.isclose(distance, 0, atol=1e-15)
    else:
        assert np.isclose(distance, actual_distance, atol=1e-15)


@pytest.mark.parametrize("delta", [0, -0.1, -0.49, -0.51])
def test_hex_collision_3d(delta):
    hex_1 = np.array(
        [[0, 0, 0], [1, 0, 0], [0, 1, 0], [1, 1, 0], [0, 0, 1], [1, 0, 1], [0, 1, 1], [1, 1, 1]],
        dtype=np.float64,
    )
    P0 = np.array([1.5 + delta, 1.5 + delta, 0.5], dtype=np.float64)
    P1 = np.array([2, 2, 1], dtype=np.float64)
    P2 = np.array([2, 1.25, 0.25], dtype=np.float64)
    P3 = P1 + P2 - P0
    quad_1 = np.array([P0, P1, P2, P3], dtype=np.float64)
    n = np.cross(quad_1[1] - quad_1[0], quad_1[2] - quad_1[0]) / np.linalg.norm(
        np.cross(quad_1[1] - quad_1[0], quad_1[2] - quad_1[0])
    )
    quad_2 = quad_1 + n
    hex_2 = np.zeros((8, 3), dtype=np.float64)
    hex_2[:4, :] = quad_1
    hex_2[4:, :] = quad_2
    actual_distance = np.linalg.norm(np.array([1, 1, P0[2]], dtype=np.float64) - hex_2[0])
    distance = np.linalg.norm(compute_distance_gjk(hex_1, hex_2))
    if P0[0] < 1:
        assert np.isclose(distance, 0, atol=1e-15)
    else:
        assert np.isclose(distance, actual_distance, atol=1e-15)


@pytest.mark.parametrize("delta", [1e8, 1.0, 1e-6, 1e-12])
@pytest.mark.parametrize("scale", [1000.0, 1.0, 1e-4])
@pytest.mark.parametrize("dtype", [np.float64])
def test_cube_distance(delta, scale, dtype):
    cubes = [
        scale
        * np.array(
            [
                [-1, -1, -1],
                [1, -1, -1],
                [-1, 1, -1],
                [1, 1, -1],
                [-1, -1, 1],
                [1, -1, 1],
                [-1, 1, 1],
                [1, 1, 1],
            ],
            dtype=dtype,
        )
    ]

    # Rotate cube 45 degrees around z, so that an edge faces along
    # x-axis (vertical)
    r = Rotation.from_euler("z", 45, degrees=True)
    cubes.append(r.apply(cubes[0]))

    # Rotate cube around y, so that a corner faces along the x-axis
    r = Rotation.from_euler("y", np.arctan2(1.0, np.sqrt(2)))
    cubes.append(r.apply(cubes[1]))

    # Rotate cube 45 degrees around y, so that an edge faces along
    # x-axis (horizontal)
    r = Rotation.from_euler("y", 45, degrees=True)
    cubes.append(r.apply(cubes[0]))

    # Rotate scene through an arbitrary angle
    r = Rotation.from_euler("xyz", [22, 13, -47], degrees=True)

    for c0 in range(4):
        for c1 in range(4):
            dx = cubes[c0][:, 0].max() - cubes[c1][:, 0].min()
            cube0 = cubes[c0]
            # Separate cubes along x-axis by distance delta
            cube1 = cubes[c1] + np.array([dx + delta, 0, 0])
            c0rot = r.apply(cube0)
            c1rot = r.apply(cube1)
            assert c0rot.dtype == dtype
            assert c1rot.dtype == dtype
            d = compute_distance_gjk(c0rot, c1rot)
            assert d.dtype == dtype
            distance = np.linalg.norm(d)
            assert np.isclose(distance, delta)


@pytest.mark.skip_in_parallel
@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_collision_2nd_order_triangle(dtype):
    points = np.array(
        [[0.0, 0.0], [1.0, 0.0], [0.0, 1.0], [0.65, 0.65], [0.0, 0.5], [0.5, 0.0]], dtype=dtype
    )
    cells = np.array([[0, 1, 2, 3, 4, 5]])
    domain = ufl.Mesh(element("Lagrange", "triangle", 2, shape=(2,), dtype=dtype))
    mesh = create_mesh(MPI.COMM_WORLD, cells, domain, points)

    # Sample points along an interior line of the domain. The last point
    # is outside the simplex made by the vertices.
    sample_points = np.array([[0.1, 0.3, 0.0], [0.2, 0.5, 0.0], [0.6, 0.6, 0.0]])

    # Create boundingboxtree
    tree = geometry.bb_tree(mesh, mesh.geometry.dim, padding=0.0)
    cell_candidates = geometry.compute_collisions_points(tree, sample_points)
    colliding_cells = geometry.compute_colliding_cells(mesh, cell_candidates, sample_points)
    # Check for collision
    for i in range(colliding_cells.num_nodes):
        assert len(colliding_cells.links(i)) == 1

    # Check if there is a point on the linear approximation of the
    # curved facet
    def line_through_points(p0, p1):
        return lambda x: (p1[1] - p0[1]) / (p1[0] - p0[0]) * (x - p0[0]) + p0[1]

    line_func = line_through_points(points[2], points[3])
    point = np.array([0.2, line_func(0.2), 0])

    # Point inside 2nd order geometry, outside linear approximation
    # Useful for debugging on a later stage
    # point = np.array([0.25, 0.89320760, 0])
    distance = geometry.squared_distance(mesh, mesh.topology.dim - 1, np.array([2]), point)
    assert np.isclose(distance, 0)



from mpi4py import MPI

import numpy as np
import pytest

import basix.ufl
import dolfinx
import ufl


@pytest.mark.parametrize("degree", range(1, 4))
@pytest.mark.parametrize("symmetry", [True, False])
def test_transpose(degree, symmetry):
    mesh = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, 10, 10)
    e = basix.ufl.element(
        "Lagrange",
        "triangle",
        degree,
        shape=(2, 2),
        symmetry=symmetry,
        dtype=dolfinx.default_real_type,
    )
    space = dolfinx.fem.functionspace(mesh, e)
    f = dolfinx.fem.Function(space)
    f.interpolate(lambda x: [x[0], x[1], 2 * x[1], x[0] ** 3])

    form = dolfinx.fem.form(ufl.inner(f - ufl.transpose(f), f - ufl.transpose(f)) * ufl.dx)
    assert np.isclose(dolfinx.fem.assemble_scalar(form), 0) == symmetry


def test_interpolation():
    """Test that a symmetric 3x3 2-tensor is correctly interpolated."""
    mesh = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, 10, 10)

    def tensor(x):
        mat = np.array([[0], [1], [2], [1], [3], [4], [2], [4], [5]])
        return np.broadcast_to(mat, (9, x.shape[1]))

    element = basix.ufl.element(
        "DG", mesh.basix_cell(), 0, shape=(3, 3), symmetry=False, dtype=dolfinx.default_real_type
    )
    space = dolfinx.fem.functionspace(mesh, element)
    f = dolfinx.fem.Function(space)
    f.interpolate(lambda x: tensor(x))

    symm_element = basix.ufl.element(
        "DG", mesh.basix_cell(), 0, shape=(3, 3), symmetry=True, dtype=dolfinx.default_real_type
    )
    symm_space = dolfinx.fem.functionspace(mesh, symm_element)
    symm_f = dolfinx.fem.Function(symm_space)
    symm_f.interpolate(lambda x: tensor(x))

    l2_error = dolfinx.fem.assemble_scalar(dolfinx.fem.form((f - symm_f) ** 2 * ufl.dx))
    atol = 10 * np.finfo(dolfinx.default_scalar_type).resolution
    assert np.isclose(l2_error, 0.0, atol=atol)


def test_eval():
    """Test that eval is correct for a symmetric 3x3 2-tensor is correct."""
    mesh = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, 10, 10)

    mat = np.array([0, 1, 2, 1, 3, 4, 2, 4, 5])

    def tensor(x):
        return np.broadcast_to(mat.reshape((9, 1)), (9, x.shape[1]))

    element = basix.ufl.element(
        "DG", mesh.basix_cell(), 0, shape=(3, 3), symmetry=True, dtype=dolfinx.default_real_type
    )
    space = dolfinx.fem.functionspace(mesh, element)
    f = dolfinx.fem.Function(space)
    f.interpolate(lambda x: tensor(x))
    value = f.eval([[0, 0, 0]], [0])
    atol = 10 * np.finfo(dolfinx.default_scalar_type).resolution
    assert np.allclose(value, mat, atol=atol)



# Copyright (C) 2022 Michal Habera and Garth N. Wells
#
# This file is part of DOLFINx (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from mpi4py import MPI

import numpy as np
import pytest

from dolfinx.cpp.mesh import cell_entity_type, cell_num_entities
from dolfinx.graph import adjacencylist
from dolfinx.io import distribute_entity_data
from dolfinx.mesh import (
    CellType,
    create_unit_cube,
    entities_to_geometry,
    locate_entities,
    meshtags_from_entities,
)
from ufl import Measure

celltypes_3D = [CellType.tetrahedron, CellType.hexahedron]


@pytest.mark.parametrize("cell_type", celltypes_3D)
def test_create(cell_type):
    comm = MPI.COMM_WORLD
    mesh = create_unit_cube(comm, 6, 6, 6, cell_type)

    marked_lines = locate_entities(mesh, 1, lambda x: np.isclose(x[1], 0.5))
    f_v = mesh.topology.connectivity(1, 0).array.reshape(-1, 2)

    entities = adjacencylist(f_v[marked_lines])
    values = np.full(marked_lines.shape[0], 2, dtype=np.int32)
    mt = meshtags_from_entities(mesh, 1, entities, values)

    assert hasattr(mt, "ufl_id")
    assert mt.indices.shape == marked_lines.shape
    assert mt.values.dtype == np.int32
    assert mt.values.shape[0] == entities.num_nodes


def test_ufl_id():
    """Test that UFL can process MeshTags (tests ufl_id attribute)."""
    comm = MPI.COMM_WORLD
    msh = create_unit_cube(comm, 6, 6, 6)
    tdim = msh.topology.dim
    marked_facets = locate_entities(msh, tdim - 1, lambda x: np.isclose(x[1], 1))
    f_v = msh.topology.connectivity(tdim - 1, 0).array.reshape(-1, 3)

    entities = adjacencylist(f_v[marked_facets])
    values = np.full(marked_facets.shape[0], 2, dtype=np.int32)
    ft = meshtags_from_entities(msh, tdim - 1, entities, values)
    ds = Measure("ds", domain=msh, subdomain_data=ft, subdomain_id=(2, 3))
    a = 1 * ds
    assert isinstance(a.subdomain_data(), dict)


def test_distribute_entity_data():
    comm = MPI.COMM_WORLD

    msh = create_unit_cube(comm, 6, 6, 6, cell_type=CellType.tetrahedron)

    def top(x):
        return np.isclose(x[2], 1)

    def bottom(x):
        return np.isclose(x[2], 0)

    # Mark some facets with unique indices
    top_facets = locate_entities(msh, msh.topology.dim - 1, top)
    bottom_facets = locate_entities(msh, msh.topology.dim - 1, bottom)
    top_values = np.full(top_facets.shape[0], 1, dtype=np.int32)
    bottom_values = np.full(bottom_facets.shape[0], 2, dtype=np.int32)

    # Concatenate and sort
    stacked_facets = np.hstack([top_facets, bottom_facets])
    stacked_values = np.hstack([top_values, bottom_values])
    facet_sort = np.argsort(stacked_facets)
    sorted_facets = stacked_facets[facet_sort]
    sorted_values = stacked_values[facet_sort]

    # Convert local facet indices into facets defined by their global
    # vertex indices
    msh.topology.create_connectivity(msh.topology.dim - 1, 0)
    facets_as_vertices = []
    for facet in sorted_facets:
        facets_as_vertices.append(msh.topology.connectivity(msh.topology.dim - 1, 0).array[facet])
    msh.topology.create_connectivity(msh.topology.dim - 1, msh.topology.dim)
    geom_vertices = entities_to_geometry(msh, msh.topology.dim - 1, np.array(sorted_facets))
    global_input = msh.geometry.input_global_indices[geom_vertices]

    # Gather all global vertices on rank 0 and distribute from there
    input_facets_as_vertices = msh.comm.gather(global_input, root=0)
    input_values = msh.comm.gather(sorted_values, root=0)

    facet_type = cell_entity_type(msh.topology.cell_type, msh.topology.dim - 1, 0)
    num_vertices_per_facet = cell_num_entities(facet_type, 0)
    if msh.comm.rank == 0:
        input_facets_as_vertices = np.vstack(input_facets_as_vertices)
        input_values = np.hstack(input_values)
    else:
        input_facets_as_vertices = np.empty((0, num_vertices_per_facet), dtype=np.int64)
        input_values = np.empty(0, dtype=np.int32)

    new_local_entities, new_values = distribute_entity_data(
        msh, msh.topology.dim - 1, input_facets_as_vertices, input_values
    )

    # Create meshtag from local entities
    new_mt = meshtags_from_entities(
        msh, msh.topology.dim - 1, adjacencylist(new_local_entities), new_values
    )

    np.testing.assert_allclose(new_mt.indices, sorted_facets)
    np.testing.assert_allclose(new_mt.values, sorted_values)



# Copyright (C) 2017-2024 Chris N. Richardson, Garth N. Wells and
# Jørgen S. Dokken
#
# This file is part of DOLFINx (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""Creation, refining and marking of meshes."""

from __future__ import annotations

import typing
import warnings
from collections.abc import Callable, Sequence
from functools import singledispatch

from mpi4py import MPI as _MPI

import numpy as np
import numpy.typing as npt

import basix
import basix.ufl
import ufl
from dolfinx import cpp as _cpp
from dolfinx import default_real_type
from dolfinx.common import IndexMap as _IndexMap
from dolfinx.cpp.mesh import (
    CellType,
    DiagonalType,
    GhostMode,
    build_dual_graph,
    cell_dim,
    to_string,
    to_type,
)
from dolfinx.cpp.refinement import (
    IdentityPartitionerPlaceholder,
    RefinementOption,
    mark_maximum,
)
from dolfinx.cpp.refinement import (
    uniform_refine as _uniform_refine,
)
from dolfinx.fem import CoordinateElement as _CoordinateElement
from dolfinx.fem import coordinate_element as _coordinate_element
from dolfinx.graph import AdjacencyList
from dolfinx.typing import Real

__all__ = [
    "CellType",
    "DiagonalType",
    "EntityMap",
    "Geometry",
    "GhostMode",
    "IdentityPartitionerPlaceholder",
    "Mesh",
    "MeshTags",
    "RefinementOption",
    "Topology",
    "build_dual_graph",
    "cell_dim",
    "compute_incident_entities",
    "compute_midpoints",
    "create_box",
    "create_cell_partitioner",
    "create_geometry",
    "create_interval",
    "create_mesh",
    "create_point_mesh",
    "create_rectangle",
    "create_submesh",
    "create_unit_cube",
    "create_unit_interval",
    "create_unit_square",
    "entities_to_geometry",
    "exterior_facet_indices",
    "locate_entities",
    "locate_entities_boundary",
    "mark_maximum",
    "meshtags",
    "meshtags_from_entities",
    "refine",
    "to_string",
    "to_type",
    "transfer_meshtag",
    "transfer_meshtags_to_submesh",
    "uniform_refine",
]


@singledispatch
def create_cell_partitioner(
    part: Callable, mode: GhostMode, max_facet_to_cell_links: int
) -> Callable:
    """Create a function to partition a mesh.

    Args:
        part: Partition function.
        mode: Ghosting mode to use.
        max_facet_to_cell_links: Maximum number of cells connected to a
            facet. Equal to 2 for non-branching manifold meshes.
            ``None`` corresponds to no upper bound on the number of
            possible connections.

    Return:
        Partitioning function.
    """
    return _cpp.mesh.create_cell_partitioner(part, mode, max_facet_to_cell_links)


@create_cell_partitioner.register(GhostMode)
def _(mode: GhostMode, max_facet_to_cell_links: int) -> Callable:
    """Create a function to partition a mesh.

    Args:
        mode: Ghosting mode to use
        max_facet_to_cell_links: Maximum number of cells connected to a
            facet. Equal to 2 for non-branching manifold meshes.
            ``None`` corresponds to no upper bound on the number of
            possible connections.

    Return:
        Partitioning function.
    """
    return _cpp.mesh.create_cell_partitioner(mode, max_facet_to_cell_links)


class Topology:
    """Topology for a :class:`dolfinx.mesh.Mesh`."""

    _cpp_object: _cpp.mesh.Topology

    def __init__(self, topology: _cpp.mesh.Topology):
        """Initialize a topology from a C++ topology.

        Args:
            topology: The C++ topology object

        Note:
            Topology objects should usually be constructed with the
            :func:`dolfinx.cpp.mesh.create_topology` and not this class
            initializer.
        """
        self._cpp_object = topology

    def cell_name(self) -> str:
        """String representation of the cell-type of the topology."""
        return to_string(self._cpp_object.cell_type)

    def connectivity(self, d0: int, d1: int) -> _cpp.graph.AdjacencyList_int32:
        """Return connectivity.

        Connectivity from entities of dimension ``d0`` to entities of
        dimension ``d1``.

        Args:
            d0: Dimension of entity one is mapping from.
            d1: Dimension of entity one is mapping to.
        """
        if (conn := self._cpp_object.connectivity(d0, d1)) is not None:
            return conn
        else:
            raise RuntimeError(
                f"Connectivity between dimension {d0} and {d1} has not been computed.",
                f"Please call `dolfinx.mesh.Topology.create_connectivity({d0}, {d1})` first.",
            )

    @property
    def comm(self):
        """Return the MPI communicator associated with the topology."""
        return self._cpp_object.comm

    def create_connectivity(self, d0: int, d1: int):
        """Build entity connectivity ``d0 -> d1``.

        Args:
            d0: Dimension of entities connectivity is from.
            d1: Dimension of entities connectivity is to.
        """
        self._cpp_object.create_connectivity(d0, d1)

    def create_entities(self, dim: int, num_threads: int = 0) -> bool:
        """Create entities of given topological dimension.

        Args:
            dim: Topological dimension of entities to create.
            num_threads: Number of CPU threads to use when creating. If
                0, threads are not spawned.

        Returns:
            ``True` is entities are created, ``False`` is if entities
            already existed.
        """
        return self._cpp_object.create_entities(dim, num_threads)

    def create_entity_permutations(self):
        """Compute entity permutations and reflections."""
        self._cpp_object.create_entity_permutations()

    @property
    def dim(self) -> int:
        """Return the topological dimension of the mesh."""
        return self._cpp_object.dim

    @property
    def entity_types(self) -> list[list[CellType]]:
        """Entity types in the topology for all topological dimensions."""
        return self._cpp_object.entity_types

    def get_cell_permutation_info(self) -> npt.NDArray[np.uint32]:
        """Get permutation information.

        The returned data is used for packing coefficients and
        assembling of tensors. The bits of each integer encodes the
        number of reflections and permutations for each sub-entity of
        the cell to be able to map it to the reference element.
        """
        return self._cpp_object.get_cell_permutation_info()

    def get_facet_permutations(self) -> npt.NDArray[np.uint8]:
        """Get the permutation integer to apply to facets.

        The bits of each integer describes the number of reflections and
        rotations that has to be applied to a facet to map between a
        facet in the mesh (relative to a cell) and the corresponding
        facet on the reference element. The data has the shape
        ``(num_cells, num_facets_per_cell)``, flattened row-wise. The
        number of cells include potential ghost cells.

        Note:
            The data can be unpacked with ``numpy.unpack_bits``.
        """
        return self._cpp_object.get_facet_permutations()

    def index_map(self, dim: int) -> _cpp.common.IndexMap:
        """Index map for the parallel distribution of the mesh entities.

        Args:
            dim: Topological dimension.

        Returns:
            Index map for the entities of dimension ``dim``.
        """
        return self._cpp_object.index_map(dim)

    def index_maps(self, dim: int) -> list[_cpp.common.IndexMap]:
        """Index maps for parallel distribution of the mesh entities.

        Args:
            dim: Topological dimension of the entities.

        Returns:
            Index maps for the entities of dimension ``dim``. May be
            empty if not yet computed.
        """
        return self._cpp_object.index_maps(dim)

    def interprocess_facets(self) -> npt.NDArray[np.int32]:
        """List of inter-process facets.

        Empty if facet topology has not been computed.
        """
        return self._cpp_object.interprocess_facets()

    @property
    def original_cell_index(self) -> npt.NDArray[np.int64]:
        """Original cell index."""
        return self._cpp_object.original_cell_index

    @property
    def cell_type(self) -> CellType:
        """Get the cell type of the topology."""
        return self._cpp_object.cell_type


class Geometry(typing.Generic[Real]):
    """The geometry of a :class:`dolfinx.mesh.Mesh`."""

    _cpp_object: _cpp.mesh.Geometry_float32 | _cpp.mesh.Geometry_float64

    def __init__(self, geometry: _cpp.mesh.Geometry_float32 | _cpp.mesh.Geometry_float64):
        """Initialize a geometry from a C++ geometry.

        Args:
            geometry: The C++ geometry object.

        Note:
            Geometry objects should usually be constructed with the
            :func:`create_geometry` and not using this class
            initializer. This class is combined with different base
            classes that depend on the scalar type used in the Geometry.
        """
        self._cpp_object = geometry

    @property
    def cmaps(self) -> list[_CoordinateElement]:
        """The coordinate maps."""
        return [_CoordinateElement(cm) for cm in self._cpp_object.cmaps]

    @property
    def cmap(self) -> _CoordinateElement:
        """The coordinate map (deprecated, use ``cmaps[0]``)."""
        warnings.warn(
            "cmap is deprecated. Use cmaps[0] instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.cmaps[0]

    @property
    def dim(self):
        """Dimension of the Euclidean coordinate system."""
        return self._cpp_object.dim

    @property
    def dofmaps(self) -> list[npt.NDArray[np.int32]]:
        """The geometry dofmaps, one per cell type."""
        return self._cpp_object.dofmaps

    @property
    def dofmap(self) -> npt.NDArray[np.int32]:
        """Dofmap for the geometry (deprecated, use ``dofmaps[0]``).

        Shape is ``(num_cells, dofs_per_cell)``.
        """
        warnings.warn(
            "dofmap is deprecated. Use dofmaps[0] instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._cpp_object.dofmaps[0]

    def index_map(self) -> _IndexMap:
        """Index map for the geometry points (nodes) distribution."""
        return self._cpp_object.index_map()

    @property
    def input_global_indices(self) -> npt.NDArray[np.int64]:
        """Global input indices of the geometry nodes."""
        return self._cpp_object.input_global_indices

    @property
    def x(self) -> npt.NDArray[Real]:
        """Geometry coordinate points.

        Shape is ``shape=(num_points, 3)``.
        """
        return self._cpp_object.x


class Mesh(typing.Generic[Real]):
    """A mesh."""

    _mesh: _cpp.mesh.Mesh_float32 | _cpp.mesh.Mesh_float64
    _topology: Topology
    _geometry: Geometry[Real]
    _ufl_domain: ufl.Mesh | None

    def __init__(
        self,
        msh: _cpp.mesh.Mesh_float32 | _cpp.mesh.Mesh_float64,
        domain: ufl.Mesh | None,
    ):
        """Initialize mesh from a C++ mesh.

        Args:
            msh: A C++ mesh object.
            domain: A UFL domain.

        Note:
            Mesh objects should usually be constructed using
            :func:`create_mesh` and not using this class initializer.
            This class is combined with different base classes that
            depend on the scalar type used in the Mesh.
        """
        self._cpp_object = msh
        self._topology = Topology(self._cpp_object.topology)
        self._geometry = Geometry(self._cpp_object.geometry)
        self._ufl_domain = domain
        if self._ufl_domain is not None:
            self._ufl_domain._ufl_cargo = self._cpp_object  # type: ignore

    @property
    def comm(self):
        """MPI communicator associated with the mesh."""
        return self._cpp_object.comm

    @property
    def name(self):
        """Name of the mesh."""
        return self._cpp_object.name

    @name.setter
    def name(self, value):
        self._cpp_object.name = value

    def ufl_cell(self) -> ufl.Cell:
        """Return the UFL cell type.

        Note:
            This method is required for UFL compatibility.
        """
        return ufl.Cell(self.topology.cell_name())

    def ufl_domain(self) -> ufl.Mesh | None:
        """Return the ufl domain corresponding to the mesh.

        Returns:
            The UFL domain. Is ``None`` if the domain has not been set.

        Note:
            This method is required for UFL compatibility.
        """
        return self._ufl_domain

    def basix_cell(self) -> basix.CellType:
        """Return the Basix cell type."""
        return getattr(basix.CellType, self.topology.cell_name())

    def h(self, dim: int, entities: npt.NDArray[np.int32]) -> npt.NDArray[np.float64]:
        """Geometric size measure of cell entities.

        Args:
            dim: Topological dimension of the entities to compute the
                size measure of.
            entities: Indices of entities of dimension ``dim`` to
                compute size measure of.

        Returns:
            Size measure for each requested entity.
        """
        return _cpp.mesh.h(self._cpp_object, dim, entities)

    @property
    def topology(self) -> Topology:
        """Mesh topology."""
        return self._topology

    @property
    def geometry(self) -> Geometry[Real]:
        """Mesh geometry."""
        return self._geometry


class MeshTags:
    """Mesh tags associate data (markers) with some mesh entities."""

    def __init__(self, meshtags):
        """Initialize tags from a C++ MeshTags object.

        Args:
            meshtags: C++ mesh tags object.

        Note:
            MeshTags objects should not usually be created using this
            initializer directly.

            A Python mesh is passed to the initializer as it may have
            UFL data attached that is not attached the C + + Mesh that is
            associated with the C + + ``meshtags`` object. If `mesh` is
            passed, ``mesh`` and ``meshtags`` must share the same C + +
            mesh.
        """
        self._cpp_object = meshtags

    def ufl_id(self) -> int:
        """Identiftying integer used by UFL."""
        return id(self)

    @property
    def topology(self) -> _cpp.mesh.Topology:
        """Mesh topology with which the tags are associated."""
        return self._cpp_object.topology

    @property
    def dim(self) -> int:
        """Topological dimension of the tagged entities."""
        return self._cpp_object.dim

    @property
    def indices(self) -> npt.NDArray[np.int32]:
        """Indices of tagged mesh entities."""
        return self._cpp_object.indices

    @property
    def values(self):
        """Values associated with tagged mesh entities."""
        return self._cpp_object.values

    @property
    def name(self) -> str:
        """Name of the mesh tags object."""
        return self._cpp_object.name

    @name.setter
    def name(self, value):
        self._cpp_object.name = value

    def find(self, value) -> npt.NDArray[np.int32]:
        """Get a list of all entity indices with a given value.

        Args:
            value: Tag value to search for.

        Returns:
            Indices of entities with tag ``value``.
        """
        return self._cpp_object.find(value)


class EntityMap:
    """Bidirectional map between entities in two topologies."""

    def __init__(self, entity_map):
        """Initialise an entity map from a C++ `EntityMap` object.

        Note:
            `EntityMap` objects should not usually be created using this
            initializer directly.

        Args:
            entity_map: A C++ `EntityMap` object.
        """
        self._cpp_object = entity_map
        self._topology = Topology(self._cpp_object.topology)
        self._sub_topology = Topology(self._cpp_object.sub_topology)

    def sub_topology_to_topology(self, entities, inverse):
        """Map entities between the sub-topology and the parent topology.

        If `inverse` is False, this function maps a list of
        `self.dim()`-dimensional entities from `self.sub_topology()` to
        the corresponding entities in `self.topology()`. If `inverse` is
        True, it performs the inverse mapping from `self.topology()` to
        `self.sub_topology()`. Entities that do not exist in the
        sub-topology are marked as -1.

        Note:
            If `inverse` is `True`, this function recomputes the inverse
            map on every call (it is not cached), which may be expensive
            if called repeatedly.

        Args:
            entities:
                A list of entity indices in the source topology.
            inverse:
                If False, maps from `self.sub_topology()` to
                `self.topology()`. If True, maps from `self.topology()`
                to `self.sub_topology()`.

        Returns:
            A list of mapped entity indices. Entities that don't exist
            in the target topology are marked as -1.
        """
        return self._cpp_object.sub_topology_to_topology(entities, inverse)

    @property
    def dim(self):
        """Topological dimension of the entities."""
        return self._cpp_object.dim

    @property
    def topology(self):
        """The topology."""
        return self._topology

    @property
    def sub_topology(self):
        """The sub-topology."""
        return self._sub_topology


def entity_map(
    topology: Topology,
    sub_topology: Topology,
    dim: int,
    sub_topology_to_topology: npt.NDArray[np.int32],
) -> EntityMap:
    """Create a bidirectional map between (sub) topologies.

    The map relates entities of dimension `dim` in `topology` and
    `sub_topology`.

    Args:
        topology: A topology.
        sub_topology: Topology of another mesh. This must be a
            "sub-topology" of `topology` i.e. every entity in
            `sub_topology` must also exist in `topology`.
        dim: The dimension of the entities
        sub_topology_to_topology: A list of entities in `topology` where
            `sub_topology_to_topology[i]` is the index in `topology`
            corresponding to entity `i` in `sub_topology`.
    """
    return EntityMap(
        _cpp.mesh.EntityMap(
            topology._cpp_object, sub_topology._cpp_object, dim, sub_topology_to_topology
        )
    )


def compute_incident_entities(
    topology: Topology, entities: npt.NDArray[np.int32], d0: int, d1: int
) -> npt.NDArray[np.int32]:
    """Compute incident entities.

    Computes entities of dimension ``d1`` connected to ``entities`` of
    dimension ``d0``.

    Args:
        topology: The topology.
        entities: List of entities of dimension ``d0``.
        d0: Topological dimension.
        d1: Topological dimension to map to.

    Returns:
        Incident entity indices.
    """
    return _cpp.mesh.compute_incident_entities(topology._cpp_object, entities, d0, d1)


def compute_midpoints(msh: Mesh, dim: int, entities: npt.NDArray[np.int32]):
    """Compute the midpoints of a set of mesh entities.

    Args:
        msh: The mesh.
        dim: Topological dimension of the mesh entities to consider.
        entities: Indices of entities in ``mesh`` to consider.

    Returns:
        Midpoints of the entities, shape ``(num_entities, 3)``.
    """
    return _cpp.mesh.compute_midpoints(msh._cpp_object, dim, entities)


def locate_entities(msh: Mesh, dim: int, marker: Callable) -> np.ndarray:
    """Compute mesh entities satisfying a geometric marking function.

    Args:
        msh: Mesh to locate entities on.
        dim: Topological dimension of the mesh entities to consider.
        marker: A function that takes an array of points ``x`` with
            shape ``(gdim, num_points)`` and returns an array of
            booleans of length ``num_points``, evaluating to `True` for
            entities to be located.

    Returns:
        Indices (local to the process) of marked mesh entities.
    """
    return _cpp.mesh.locate_entities(msh._cpp_object, dim, marker)


def locate_entities_boundary(msh: Mesh, dim: int, marker: Callable) -> np.ndarray:
    """Mesh entities connected to a boundary facet a geometric condition.

    Compute indices of all mesh entities that are attached to an owned
    boundary facet and evaluate to true for the provided geometric
    marking function. An entity is considered marked if the marker
    function evaluates to true for all of its vertices.

    For vertices and edges, in parallel this function will not
    necessarily mark all entities that are on the exterior boundary. For
    example, it is possible for a process to have a vertex that lies on
    the boundary without any of the attached facets being a boundary
    facet. When used to find degrees-of-freedom, e.g. using
    :func:`dolfinx.fem.locate_dofs_topological`, the function that uses
    the data returned by this function must typically perform some
    parallel communication.

    Args:
        msh: Mesh to locate boundary entities on.
        dim: Topological dimension of the mesh entities to consider
        marker: Function that takes an array of points ``x`` with shape
            ``(gdim, num_points)`` and returns an array of booleans of
            length ``num_points``, evaluating to ``True`` for entities
            to be located.

    Returns:
        Indices (local to the process) of marked mesh entities.
    """
    return _cpp.mesh.locate_entities_boundary(msh._cpp_object, dim, marker)


def transfer_meshtag(
    meshtag: MeshTags,
    msh1: Mesh,
    parent_cell: npt.NDArray[np.int32],
    parent_facet: npt.NDArray[np.int8] | None = None,
) -> MeshTags:
    """Create cell mesh tags on a refined mesh from parent mesh tags.

    Args:
        meshtag: Mesh tags on the coarse, parent mesh.
        msh1: The refined mesh.
        parent_cell: Index of the parent cell for each cell in the
            refined mesh.
        parent_facet: Index of the local parent facet for each cell
            in the refined mesh. Only required for transfer tags on
            facets.

    Returns:
        Mesh tags on the refined mesh.
    """
    if meshtag.dim == meshtag.topology.dim:
        mt = _cpp.refinement.transfer_cell_meshtag(
            meshtag._cpp_object, msh1.topology._cpp_object, parent_cell
        )
        return MeshTags(mt)
    elif meshtag.dim == meshtag.topology.dim - 1:
        assert parent_facet is not None
        mt = _cpp.refinement.transfer_facet_meshtag(
            meshtag._cpp_object, msh1.topology._cpp_object, parent_cell, parent_facet
        )
        return MeshTags(mt)
    else:
        raise RuntimeError("MeshTag transfer is supported on on cells or facets.")


def uniform_refine(
    msh: Mesh,
    partitioner: Callable | None = None,
) -> Mesh:
    """Uniformly refine a mesh.

    Note:
        Using the default partitioner for the refined mesh, the refined
        mesh will **not** include ghosts cells (cells connected by facet
        to an owned cells) even if the parent mesh is ghosted. The
        refined cells will be on the same process as the parent cell.

    Args:
        msh: Mesh from which to create the refined mesh.
        partitioner: Partitioner to distribute the refined mesh.
            If ``None`` is passed (default) no redistribution will happen.

    Returns:
        The refined mesh.
    """
    _cpp_mesh = _uniform_refine(msh._cpp_object, partitioner)
    # Create new ufl domain as it will carry a reference to the C++ mesh
    # in the ufl_cargo
    ufl_domain = ufl.Mesh(msh._ufl_domain.ufl_coordinate_element())  # type: ignore
    return Mesh(_cpp_mesh, ufl_domain)


def refine(
    msh: Mesh,
    edges: np.ndarray | None = None,
    partitioner: Callable | IdentityPartitionerPlaceholder = IdentityPartitionerPlaceholder(),
    option: RefinementOption = RefinementOption.parent_cell,
) -> tuple[Mesh, npt.NDArray[np.int32], npt.NDArray[np.int8]]:
    """Refine a mesh.

    Note:
        Using the `None` partitioner for the refined mesh, the refined mesh
        will **not** include ghosts cells (cells connected by facet to an
        owned cells) even if the parent mesh is ghosted. The refined cells
        will be on the same process as the parent cell.

    Args:
        msh: Mesh from which to create the refined mesh.
        edges: Indices of edges to split during refinement. If ``None``,
            mesh refinement is uniform.
        partitioner: Partitioner to distribute the refined mesh. If a
            :py:class:`IdentityPartitionerPlaceholder` is passed (default)
            no redistribution is performed, i.e. refined cells remain on
            the same process as the parent cell, but the ghost layer is
            updated. If a custom partitioner is passed, it will be used for
            distributing the refined mesh. If ``None`` is passed no
            redistribution will happen.
        option: Controls whether parent cells and/or parent facets are
            computed.

    Returns:
       Refined mesh, (optional) parent cells, (optional) parent facets
    """
    mesh1, parent_cell, parent_facet = _cpp.refinement.refine(
        msh._cpp_object, edges, partitioner, option
    )
    # Create new ufl domain as it will carry a reference to the C++ mesh
    # in the ufl_cargo
    ufl_domain = ufl.Mesh(msh._ufl_domain.ufl_coordinate_element())  # type: ignore
    return Mesh(mesh1, ufl_domain), parent_cell, parent_facet


def create_mesh(
    comm: _MPI.Comm,
    cells: npt.NDArray[np.int64],
    e: ufl.Mesh | basix.finite_element.FiniteElement | basix.ufl._BasixElement | _CoordinateElement,
    x: npt.NDArray[np.floating],
    partitioner: Callable | None = None,
    max_facet_to_cell_links: int = 2,
) -> Mesh:
    """Create a mesh from topology and geometry arrays.

    Args:
        comm: MPI communicator to define the mesh on.
        cells: Cells of the mesh. ``cells[i]`` are the 'nodes' of
            cell ``i``.
        e: UFL mesh. The mesh scalar type is determined by the scalar
            type of ``e``.
        x: Mesh geometry ('node' coordinates), with shape
            ``(num_nodes, gdim)``.
        partitioner: Function that determines the parallel distribution of
            cells across MPI ranks.
        max_facet_to_cell_links: Maximum number of cells a facet can
            be connected to.

    Note:
        If required, the coordinates ``x`` will be cast to the same
        scalar type as the domain/element ``e``.

    Returns:
        A mesh.
    """
    if partitioner is None and comm.size > 1:
        partitioner = create_cell_partitioner(GhostMode.none, 2)  # type: ignore

    x = np.asarray(x, order="C")
    if x.ndim == 1:
        gdim = 1
    else:
        gdim = x.shape[1]

    dtype = None
    if isinstance(e, ufl.domain.Mesh):
        # e is a UFL domain
        e_ufl = e.ufl_coordinate_element()  # type: ignore
        cmap = _coordinate_element(e_ufl.basix_element)  # type: ignore
        domain = e
        dtype = cmap.dtype
        # TODO: Resolve UFL vs Basix geometric dimension issue
        # assert domain.geometric_dimension() == gdim
    elif isinstance(e, basix.finite_element.FiniteElement):
        # e is a Basix element
        # TODO: Resolve geometric dimension vs shape for manifolds
        cmap = _coordinate_element(e)  # type: ignore
        e_ufl = basix.ufl._BasixElement(e)  # type: ignore
        e_ufl = basix.ufl.blocked_element(e_ufl, shape=(gdim,))
        domain = ufl.Mesh(e_ufl)
        dtype = cmap.dtype
        assert domain.geometric_dimension == gdim
    elif isinstance(e, ufl.finiteelement.AbstractFiniteElement):
        # e is a Basix 'UFL' element
        cmap = _coordinate_element(e.basix_element)  # type: ignore
        domain = ufl.Mesh(e)
        dtype = cmap.dtype
        assert domain.geometric_dimension == gdim
    elif isinstance(e, _CoordinateElement):
        # e is a CoordinateElement
        cmap = e
        domain = None
        dtype = cmap.dtype  # type: ignore
    else:
        raise ValueError(f"Unsupported element type {type(e)}.")

    x = np.asarray(x, dtype=dtype, order="C")
    cells = np.asarray(cells, dtype=np.int64, order="C")
    msh: _cpp.mesh.Mesh_float32 | _cpp.mesh.Mesh_float64 = _cpp.mesh.create_mesh(
        comm, cells, cmap._cpp_object, x, partitioner, max_facet_to_cell_links
    )

    return Mesh(msh, domain)  # type: ignore


def create_submesh(
    msh: Mesh, dim: int, entities: npt.NDArray[np.int32]
) -> tuple[Mesh, EntityMap, EntityMap, npt.NDArray[np.int32]]:
    """Create a mesh with specified entities from another mesh.

    Args:
        msh: Mesh to create the sub-mesh from.
        dim: Topological dimension of the entities in ``msh`` to include
            in the sub-mesh.
        entities: Indices of entities in ``msh`` to include in the
            sub-mesh.

    Returns:
        The (1) sub mesh, (2) entity map, (3) vertex map, and (4) node
        map (geometry). Each of the maps maps a local index of the sub mesh
        to a local index of ``msh``.
    """
    submsh, entity_map, vertex_map, geom_map = _cpp.mesh.create_submesh(
        msh._cpp_object, dim, entities
    )
    submsh_domain = ufl.Mesh(
        basix.ufl.element(
            "Lagrange",
            to_string(submsh.topology.cell_type),
            submsh.geometry.cmaps[0].degree,
            basix.LagrangeVariant(submsh.geometry.cmaps[0].variant),
            shape=(submsh.geometry.dim,),
            dtype=submsh.geometry.x.dtype,
        )
    )
    return (Mesh(submsh, submsh_domain), EntityMap(entity_map), EntityMap(vertex_map), geom_map)


def meshtags(
    msh: Mesh,
    dim: int,
    entities: npt.NDArray[np.int32],
    values: np.ndarray | int | float,
) -> MeshTags:
    """Create mesh tags that associate data with a subset of mesh entities.

    Args:
        msh: The mesh.
        dim: Topological dimension of the mesh entity.
        entities: Indices(local to process) of entities to associate
            values with . The array must be sorted and must not contain
            duplicates.
        values: The corresponding value for each entity.

    Returns:
        A mesh tags object.

    Note:
        The type of the returned MeshTags is inferred from the type of
        ``values``.
    """
    if isinstance(values, int):
        assert values >= np.iinfo(np.int32).min and values <= np.iinfo(np.int32).max
        values = np.full(entities.shape, values, dtype=np.int32)
    elif isinstance(values, float):
        values = np.full(entities.shape, values, dtype=np.double)

    values = np.asarray(values)
    if values.dtype == np.int8:
        ftype = _cpp.mesh.MeshTags_int8
    elif values.dtype == np.int32:
        ftype = _cpp.mesh.MeshTags_int32
    elif values.dtype == np.int64:
        ftype = _cpp.mesh.MeshTags_int64
    elif values.dtype == np.float64:
        ftype = _cpp.mesh.MeshTags_float64
    else:
        raise NotImplementedError(f"Type {values.dtype} not supported.")

    return MeshTags(
        ftype(msh.topology._cpp_object, dim, np.asarray(entities, dtype=np.int32), values)
    )


def meshtags_from_entities(
    msh: Mesh, dim: int, entities: AdjacencyList, values: npt.NDArray[typing.Any]
) -> MeshTags:
    """Create mesh tags that associate data with a subset of mesh entities.

    The entities are defined by their vertices.

    Args:
        msh: The mesh.
        dim: Topological dimension of the mesh entity.
        entities: Entities to associated values with, with entities
            defined by their vertices.
        values: The corresponding value for each entity.

    Returns:
        A mesh tags object.

    Note:
        The type of the returned MeshTags is inferred from the type of
        ``values``.
    """
    if isinstance(values, int):
        assert np.can_cast(values, np.int32)
        values = np.full(entities.num_nodes, values, dtype=np.int32)
    elif isinstance(values, float):
        values = np.full(entities.num_nodes, values, dtype=np.double)
    values = np.asarray(values)
    return MeshTags(
        _cpp.mesh.create_meshtags(msh.topology._cpp_object, dim, entities._cpp_object, values)
    )


def create_interval(
    comm: _MPI.Comm,
    nx: int,
    points: npt.ArrayLike,
    dtype: npt.DTypeLike = default_real_type,
    ghost_mode=GhostMode.shared_facet,
    partitioner=None,
    gdim: int = 1,
) -> Mesh:
    """Create an interval mesh.

    Args:
        comm: MPI communicator.
        nx: Number of cells.
        points: Coordinates of the end points.
        dtype: Float type for the mesh geometry(``numpy.float32``
            or ``numpy.float64``).
        ghost_mode: Ghost mode used in the mesh partitioning. Options
            are ``GhostMode.none`` and ``GhostMode.shared_facet``.
        partitioner: Partitioning function to use for determining the
            parallel distribution of cells across MPI ranks.
        gdim: Geometric dimension. The interval lies along the first
            coordinate axis; remaining components are zero.

    Returns:
        An interval mesh.
    """
    if partitioner is None and comm.size > 1:
        partitioner = _cpp.mesh.create_cell_partitioner(ghost_mode, 2)
    domain = ufl.Mesh(
        basix.ufl.element(
            "Lagrange",
            "interval",
            1,
            lagrange_variant=basix.LagrangeVariant.unset,
            shape=(gdim,),
            dtype=dtype,
        )
    )  # type: ignore
    if np.issubdtype(dtype, np.float32):
        msh = _cpp.mesh.create_interval_float32(comm, nx, points, ghost_mode, partitioner, gdim)
    elif np.issubdtype(dtype, np.float64):
        msh = _cpp.mesh.create_interval_float64(comm, nx, points, ghost_mode, partitioner, gdim)
    else:
        raise RuntimeError(f"Unsupported mesh geometry float type: {dtype}")

    return Mesh(msh, domain)


def create_unit_interval(
    comm: _MPI.Comm,
    nx: int,
    dtype: npt.DTypeLike = default_real_type,
    ghost_mode=GhostMode.shared_facet,
    partitioner=None,
    gdim: int = 1,
) -> Mesh:
    """Create a mesh on the unit interval.

    Args:
        comm: MPI communicator.
        nx: Number of cells.
        dtype: Float type for the mesh geometry(``numpy.float32``
            or ``numpy.float64``).
        ghost_mode: Ghost mode used in the mesh partitioning. Options
            are ``GhostMode.none`` and ``GhostMode.shared_facet``.
        partitioner: Partitioning function to use for determining the
            parallel distribution of cells across MPI ranks.
        gdim: Geometric dimension. The interval lies along the first
            coordinate axis; remaining components are zero.

    Returns:
        A unit interval mesh with end points at 0 and 1.
    """
    return create_interval(
        comm, nx, [0.0, 1.0], dtype=dtype, ghost_mode=ghost_mode, partitioner=partitioner, gdim=gdim
    )


def create_rectangle(
    comm: _MPI.Comm,
    points: npt.ArrayLike,
    n: Sequence[int],
    cell_type=CellType.triangle,
    dtype: npt.DTypeLike = default_real_type,
    ghost_mode=GhostMode.shared_facet,
    partitioner=None,
    diagonal: DiagonalType = DiagonalType.right,
    gdim: int = 2,
) -> Mesh:
    """Create a rectangle mesh.

    Args:
        comm: MPI communicator.
        points: Coordinates of the lower - left and upper - right
            corners of the rectangle.
        n: Number of cells in each direction.
        cell_type: Mesh cell type.
        dtype: Float type for the mesh geometry(``numpy.float32``
            or ``numpy.float64``)
        ghost_mode: Ghost mode used in the mesh partitioning.
        partitioner: Function that computes the parallel distribution of
            cells across MPI ranks.
        diagonal: Direction of diagonal of triangular meshes.
            See :class:`DiagonalType` for available options.
        gdim: Geometric dimension. The rectangle lies in the first 2
            coordinate axes; remaining components are zero.

    Returns:
        A mesh of a rectangle.
    """
    if partitioner is None and comm.size > 1:
        partitioner = _cpp.mesh.create_cell_partitioner(ghost_mode, 2)
    domain = ufl.Mesh(
        basix.ufl.element(
            "Lagrange",
            cell_type.name,
            1,
            lagrange_variant=basix.LagrangeVariant.unset,
            shape=(gdim,),
            dtype=dtype,
        )
    )  # type: ignore
    if np.issubdtype(dtype, np.float32):
        msh = _cpp.mesh.create_rectangle_float32(
            comm, points, n, cell_type, partitioner, diagonal, gdim
        )
    elif np.issubdtype(dtype, np.float64):
        msh = _cpp.mesh.create_rectangle_float64(
            comm, points, n, cell_type, partitioner, diagonal, gdim
        )
    else:
        raise RuntimeError(f"Unsupported mesh geometry float type: {dtype}")

    return Mesh(msh, domain)


def create_unit_square(
    comm: _MPI.Comm,
    nx: int,
    ny: int,
    cell_type=CellType.triangle,
    dtype: npt.DTypeLike = default_real_type,
    ghost_mode=GhostMode.shared_facet,
    partitioner=None,
    diagonal: DiagonalType = DiagonalType.right,
    gdim: int = 2,
) -> Mesh:
    """Create a mesh of a unit square.

    Args:
        comm: MPI communicator.
        nx: Number of cells in the "x" direction.
        ny: Number of cells in the "y" direction.
        cell_type: Mesh cell type.
        dtype: Float type for the mesh geometry(``numpy.float32``
            or ``numpy.float64``).
        ghost_mode: Ghost mode used in the mesh partitioning.
        partitioner: Function that computes the parallel distribution of
            cells across MPI ranks.
        diagonal:
            Direction of diagonal. See :class:`DiagonalType` for
            available options.
        gdim: Geometric dimension. The square lies in the first 2
            coordinate axes; remaining components are zero.

    Returns:
        A mesh of a square with corners at ``(0, 0)`` and ``(1, 1)``.
    """
    return create_rectangle(
        comm,
        [np.array([0.0, 0.0]), np.array([1.0, 1.0])],
        (nx, ny),
        cell_type,
        dtype=dtype,
        ghost_mode=ghost_mode,
        partitioner=partitioner,
        diagonal=diagonal,
        gdim=gdim,
    )


def create_box(
    comm: _MPI.Comm,
    points: list[npt.ArrayLike],
    n: Sequence[int],
    cell_type=CellType.tetrahedron,
    dtype: npt.DTypeLike = default_real_type,
    ghost_mode=GhostMode.shared_facet,
    partitioner=None,
) -> Mesh:
    """Create a box mesh.

    Args:
        comm: MPI communicator.
        points: Coordinates of the 'lower-left' and 'upper-right'
            corners of the box.
        n: List of cells in each direction
        cell_type: The cell type.
        dtype: Float type for the mesh geometry(``numpy.float32``
            or ``numpy.float64``).
        ghost_mode: The ghost mode used in the mesh partitioning.
        partitioner: Function that computes the parallel distribution of
            cells across MPI ranks.

    Returns:
        A mesh of a box domain.
    """
    if partitioner is None and comm.size > 1:
        partitioner = _cpp.mesh.create_cell_partitioner(ghost_mode, 2)
    domain = ufl.Mesh(
        basix.ufl.element(
            "Lagrange",
            cell_type.name,
            1,
            lagrange_variant=basix.LagrangeVariant.unset,
            shape=(3,),
            dtype=dtype,
        )
    )  # type: ignore
    if np.issubdtype(dtype, np.float32):
        msh = _cpp.mesh.create_box_float32(comm, points, n, cell_type, partitioner)
    elif np.issubdtype(dtype, np.float64):
        msh = _cpp.mesh.create_box_float64(comm, points, n, cell_type, partitioner)
    else:
        raise RuntimeError(f"Unsupported mesh geometry float type: {dtype}")

    return Mesh(msh, domain)


def create_unit_cube(
    comm: _MPI.Comm,
    nx: int,
    ny: int,
    nz: int,
    cell_type=CellType.tetrahedron,
    dtype: npt.DTypeLike = default_real_type,
    ghost_mode=GhostMode.shared_facet,
    partitioner=None,
) -> Mesh:
    """Create a mesh of a unit cube.

    Args:
        comm: MPI communicator.
        nx: Number of cells in "x" direction.
        ny: Number of cells in "y" direction.
        nz: Number of cells in "z" direction.
        cell_type: Mesh cell type
        dtype: Float type for the mesh geometry(``numpy.float32``
            or ``numpy.float64``).
        ghost_mode: Ghost mode used in the mesh partitioning.
        partitioner: Function that computes the parallel distribution of
            cells across MPI ranks.

    Returns:
        A mesh of an axis-aligned unit cube with corners at ``(0, 0, 0)``
        and ``(1, 1, 1)``.
    """
    return create_box(
        comm,
        [np.array([0.0, 0.0, 0.0]), np.array([1.0, 1.0, 1.0])],
        (nx, ny, nz),
        cell_type,
        dtype,
        ghost_mode,
        partitioner,
    )


def entities_to_geometry(
    msh: Mesh, dim: int, entities: npt.NDArray[np.int32], permute=False
) -> npt.NDArray[np.int32]:
    """Geometric DOFs associated with the closure of given mesh entities.

    Args:
        msh: The mesh.
        dim: Topological dimension of the entities of interest.
        entities: Entity indices (local to the process).
        permute: Permute the DOFs such that they are consistent with the
            orientation of `dim`-dimensional mesh entities. This
            requires `create_entity_permutations` to be called first.

    Returns:
        The geometric DOFs associated with the closure of the entities
        in `entities`.
    """
    return _cpp.mesh.entities_to_geometry(msh._cpp_object, dim, entities, permute)


def exterior_facet_indices(topology: Topology) -> npt.NDArray[np.int32]:
    """Compute the indices of exterior facets that are owned by the caller.

    An exterior facet (co-dimension 1) is one that is connected globally
    to only one cell of co-dimension 0).

    Note:
        This is a collective operation that should be called on all
        processes.

    Args:
        topology: The topology

    Returns:
        Sorted list of owned facet indices that are exterior facets of
        the mesh.
    """
    return _cpp.mesh.exterior_facet_indices(topology._cpp_object)


def create_geometry(
    index_map: _IndexMap,
    dofmap: npt.NDArray[np.int32],
    element: _CoordinateElement,
    x: np.ndarray,
    input_global_indices: npt.NDArray[np.int64],
) -> Geometry:
    """Create a Geometry object.

    Args:
        index_map: Index map describing the layout of the geometry
            points (nodes).
        dofmap: The geometry (point) dofmap. For a cell, it gives the
            row in the point coordinates ``x`` of each local geometry
            node. ``shape=(num_cells, num_dofs_per_cell)``.
        element: Element that describes the cell geometry map.
        x: The point coordinates. The shape is
            ``(num_points, geometric_dimension).``
        input_global_indices: The 'global' input index of each point,
            commonly from a mesh input file.
    """
    if x.dtype == np.float64:
        ftype = _cpp.mesh.Geometry_float64
    elif x.dtype == np.float32:
        ftype = _cpp.mesh.Geometry_float32
    else:
        raise ValueError("Unknown floating type for geometry, got: {x.dtype}")

    if (dtype := np.dtype(element.dtype)) != x.dtype:
        raise ValueError(f"Mismatch in x dtype ({x.dtype}) and coordinate element ({dtype})")
    return Geometry(ftype(index_map, dofmap, element._cpp_object, x, input_global_indices))


def transfer_meshtags_to_submesh(
    entity_tag: MeshTags,
    submesh: Mesh,
    vertex_to_parent: EntityMap,
    cell_to_parent: EntityMap,
) -> MeshTags:
    """Transfer a ``entity_tag`` from a parent mesh to a ``submesh``.

    Args:
        entity_tag: Tag to transfer
        submesh: Submesh to transfer tag to
        vertex_to_parent: Mapping from submesh vertices to parent
            mesh vertices
        cell_to_parent: Mapping from submesh cells to parent entities
    Returns:
        The transferred meshtags object on the submesh.
    """
    dim = entity_tag.dim
    sub_tdim = submesh.topology.dim
    if dim > sub_tdim:
        raise RuntimeError(
            f"Cannot transfer meshtags of dimension {dim} to submesh with topological dimension"
        )

    submesh.topology.create_connectivity(sub_tdim, sub_tdim)
    submesh.topology.create_connectivity(entity_tag.dim, 0)
    submesh.topology.create_connectivity(sub_tdim, entity_tag.dim)
    entity_tag.topology.create_connectivity(dim, 0)
    entity_tag.topology.create_connectivity(dim, sub_tdim)
    dtype = entity_tag.values.dtype
    if dtype == np.int32:
        ftype = _cpp.mesh.transfer_meshtags_to_submesh_int32
    elif dtype == np.int64:
        ftype = _cpp.mesh.transfer_meshtags_to_submesh_int64
    elif dtype == np.float64:
        ftype = _cpp.mesh.transfer_meshtags_to_submesh_float64
    else:
        raise NotImplementedError(
            f"MeshTags with dtype {dtype} not supported for transfer to submesh."
        )
    cpp_tag = ftype(
        entity_tag._cpp_object,
        submesh.topology._cpp_object,
        vertex_to_parent._cpp_object,
        cell_to_parent._cpp_object,
    )
    return MeshTags(cpp_tag)


def create_point_mesh(comm: _MPI.Intracomm, points: npt.NDArray[np.float32 | np.float64]) -> Mesh:
    """Create a mesh consisting of points only.

    Note:
        No points are shared between processes.

    Args:
        comm: MPI communicator to create the mesh on.
        points: Points local to the process in the mesh.
    """
    # Create topology which only has a 0->0 connectivity and dim 0 entities
    cells = np.arange(points.shape[0], dtype=np.int32).reshape(-1, 1)
    num_nodes_local = cells.shape[0]
    imap = _cpp.common.IndexMap(comm, num_nodes_local)
    local_range = imap.local_range[0]
    igi = np.arange(num_nodes_local, dtype=np.int64) + local_range
    topology = _cpp.mesh.Topology(
        cell_type=_cpp.mesh.CellType.point,
        vertex_map=imap,
        cell_map=imap,
        cells=_cpp.graph.AdjacencyList_int32(cells),
        original_index=igi,
    )

    e = basix.ufl.element("Lagrange", "point", 0, shape=(points.shape[1],), dtype=points.dtype)
    c_el = _coordinate_element(e.basix_element)  # type: ignore[call-arg]
    geometry = create_geometry(imap, cells, c_el, points, igi)

    if points.dtype == np.float64:
        cpp_mesh = _cpp.mesh.Mesh_float64(comm, topology, geometry._cpp_object)
    elif points.dtype == np.float32:
        cpp_mesh = _cpp.mesh.Mesh_float32(comm, topology, geometry._cpp_object)
    else:
        raise RuntimeError(f"Unsupported dtype for mesh {points.dtype}")

    return Mesh(cpp_mesh, domain=ufl.Mesh(e))



# Copyright (C) 2018-2022 Garth N. Wells, Jack S. Hale
#
# This file is part of DOLFINx (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""Assembly functions for variational forms."""

from __future__ import annotations

import functools
import typing
import warnings
from collections.abc import Sequence

import numpy as np
import numpy.typing as npt

import dolfinx
from dolfinx import cpp as _cpp
from dolfinx import default_scalar_type, la
from dolfinx.cpp.fem import pack_coefficients as _pack_coefficients
from dolfinx.cpp.fem import pack_constants as _pack_constants
from dolfinx.fem import IntegralType
from dolfinx.fem.bcs import DirichletBC
from dolfinx.fem.forms import Form
from dolfinx.fem.function import FunctionSpace


def pack_constants(
    form: Form | Sequence[Form],
) -> npt.NDArray | Sequence[npt.NDArray]:
    """Pack form constants for use in assembly.

    Pack the 'constants' that appear in forms. The packed constants can
    then be passed to an assembler. This is a performance optimisation
    for cases where a form is assembled multiple times and (some)
    constants do not change.

    If ``form`` is a sequence of forms, this function returns an array
    of form constants with the same shape as ``form``.

    Args:
        form: Single form or sequence of forms to pack the constants
            for.

    Returns:
        A ``constant`` array for each form.
    """

    def _pack(form):
        if form is None:
            return None
        elif isinstance(form, Sequence):
            return list(map(lambda sub_form: _pack(sub_form), form))
        else:
            return _pack_constants(form._cpp_object)

    return _pack(form)


def pack_coefficients(
    form: Form | Sequence[Form],
) -> (
    dict[tuple[IntegralType, int], npt.NDArray] | list[dict[tuple[IntegralType, int], npt.NDArray]]
):
    """Pack form coefficients for use in assembly.

    Pack the ``coefficients`` that appear in forms. The packed
    coefficients can be passed to an assembler. This is a performance
    optimisation for cases where a form is assembled multiple times and
    (some) coefficients do not change.

    If ``form`` is an array of forms, this function returns an array of
    form coefficients with the same shape as ``form``.

    Args:
        form: A form or a sequence of forms to pack the coefficients
        for.

    Returns:
        Coefficients for each form.
    """

    def _pack(form):
        if form is None:
            return {}
        elif isinstance(form, Sequence):
            return list(map(lambda sub_form: _pack(sub_form), form))
        else:
            return _pack_coefficients(form._cpp_object)

    return _pack(form)


# -- Vector and matrix instantiation --------------------------------------


def create_vector(V: FunctionSpace, dtype: npt.DTypeLike = default_scalar_type) -> la.Vector:
    """Create a Vector that is compatible with the given function space.

    Args:
        V: A function space.
        dtype: Data type of the vector.

    Returns:
        A vector compatible with the function space.
    """
    # Can just take the first dofmap here, since all dof maps have the same
    # index map in mixed-topology meshes
    dofmap = V.dofmaps[0]  # type: ignore[attr-defined]
    return la.vector(dofmap.index_map, dofmap.index_map_bs, dtype=dtype)


def create_matrix(a: Form, block_mode: la.BlockMode | None = None) -> la.MatrixCSR:
    """Create a sparse matrix that is compatible with a bilinear form.

    Args:
        a: Bilinear form.
        block_mode: Block mode of the CSR matrix. If ``None``, default
            is used.

    Returns:
        A sparse matrix that the form can be assembled into.
    """
    sp = dolfinx.fem.create_sparsity_pattern(a)
    sp.finalize()
    if block_mode is not None:
        return la.matrix_csr(sp, block_mode=block_mode, dtype=a.dtype)
    else:
        return la.matrix_csr(sp, dtype=a.dtype)


# -- Scalar assembly ------------------------------------------------------


def assemble_scalar(
    M: Form,
    constants: npt.NDArray | None = None,
    coeffs: dict[tuple[IntegralType, int], npt.NDArray] | None = None,
) -> float | complex:
    """Assemble functional.

    The returned value is local and not accumulated across processes.

    Args:
        M: The functional to compute.
        constants: Constants that appear in the form. If ``None``, any
            required constants will be computed.
        coeffs: Coefficients that appear in the form. If not provided,
            any required coefficients will be computed.

    Returns:
        The computed scalar on the calling rank.

    Note:
        Passing `constants` and `coefficients` is a performance
        optimisation for when a form is assembled multiple times and
        when (some) constants and coefficients are unchanged.

        To compute the functional value on the whole domain, the output
        of this function is typically summed across all MPI ranks.
    """
    constants = pack_constants(M) if constants is None else constants  # type: ignore[assignment]
    coeffs = pack_coefficients(M) if coeffs is None else coeffs  # type: ignore[assignment]
    return _cpp.fem.assemble_scalar(M._cpp_object, constants, coeffs)


# -- Vector assembly ------------------------------------------------------


@functools.singledispatch
def assemble_vector(
    L: typing.Any,
    constants: npt.NDArray | None = None,
    coeffs: dict[tuple[IntegralType, int], npt.NDArray] | None = None,
) -> la.Vector:
    """Assemble linear form into a vector."""
    return _assemble_vector_form(L, constants, coeffs)


@assemble_vector.register
def _assemble_vector_form(
    L: Form,
    constants: npt.NDArray | None = None,
    coeffs: dict[tuple[IntegralType, int], npt.NDArray] | None = None,
) -> la.Vector:
    """Assemble linear form into a new Vector.

    Args:
        L: The linear form to assemble.
        constants: Constants that appear in the form. If ``None``,
            any required constants will be computed.
        coeffs: Coefficients that appear in the form. If not provided,
            any required coefficients will be computed.

    Returns:
        The assembled vector for the calling rank.

    Note:
        Passing `constants` and `coefficients` is a performance
        optimisation for when a form is assembled multiple times and
        when (some) constants and coefficients are unchanged.

    Note:
        The returned vector is not finalised, i.e. ghost values are not
        accumulated on the owning processes. Calling
        :func:`dolfinx.la.Vector.scatter_reverse` on the return vector
        can accumulate ghost contributions.
    """
    b = create_vector(L.function_spaces[0], L.dtype)
    b.array[:] = 0
    constants = pack_constants(L) if constants is None else constants  # type: ignore[assignment]
    coeffs = pack_coefficients(L) if coeffs is None else coeffs  # type: ignore[assignment]
    _assemble_vector_array(b.array, L, constants, coeffs)
    return b


@assemble_vector.register(np.ndarray)
def _assemble_vector_array(
    b: npt.NDArray,
    L: Form,
    constants: npt.NDArray | None = None,
    coeffs: dict[tuple[IntegralType, int], npt.NDArray] | None = None,
) -> npt.NDArray:
    """Assemble linear form into an existing array.

    Args:
        b: Array to assemble the contribution from the calling MPI
            rank into. It must have the required size.
        L: Linear form assemble.
        constants: Constants that appear in the form. If ``None``,
            any required constants will be computed.
        coeffs: Coefficients that appear in the form. If not provided,
            any required coefficients will be computed.

    Note:
        Passing `constants` and `coefficients` is a performance
        optimisation for when a form is assembled multiple times and
        when (some) constants and coefficients are unchanged.

    Note:
        The returned vector is not finalised, i.e. ghost values are not
        accumulated on the owning processes. Calling
        :func:`dolfinx.la.Vector.scatter_reverse` on the return vector
        can accumulate ghost contributions.
    """
    constants = pack_constants(L) if constants is None else constants  # type: ignore[assignment]
    coeffs = pack_coefficients(L) if coeffs is None else coeffs  # type: ignore[assignment]
    _cpp.fem.assemble_vector(b, L._cpp_object, constants, coeffs)
    return b


# -- Matrix assembly ------------------------------------------------------


@functools.singledispatch
def assemble_matrix(
    a: typing.Any,
    bcs: Sequence[DirichletBC] | None = None,
    diag: float = 1.0,
    constants: npt.NDArray | None = None,
    coeffs: dict[tuple[IntegralType, int], npt.NDArray] | None = None,
    block_mode: la.BlockMode | None = None,
) -> la.MatrixCSR:
    """Assemble bilinear form into a matrix.

    Args:
        a: The bilinear form assemble.
        bcs: Boundary conditions that affect the assembled matrix.
            Degrees-of-freedom constrained by a boundary condition will
            have their rows/columns zeroed and the value ``diagonal``
            set on on the matrix diagonal.
        diag: Value to set on the matrix diagonal for Dirichlet
            boundary condition constrained degrees-of-freedom belonging
            to the same trial and test space.
        constants: Constants that appear in the form. If ``None``,
            any required constants will be computed.
        coeffs: Coefficients that appear in the form. If not provided,
            any required coefficients will be computed.
        block_mode: Block size mode for the returned space matrix. If
            ``None``, default is used.

    Returns:
        Matrix representation of the bilinear form ``a``.

    Note:
        The returned matrix is not finalised, i.e. ghost values are not
        accumulated.
    """
    bcs = [] if bcs is None else bcs
    A: la.MatrixCSR = create_matrix(a, block_mode)
    _assemble_matrix_csr(A, a, bcs, diag, constants, coeffs)
    return A


@assemble_matrix.register
def _assemble_matrix_csr(
    A: la.MatrixCSR,
    a: Form,
    bcs: Sequence[DirichletBC] | None = None,
    diag: float = 1.0,
    constants: npt.NDArray | None = None,
    coeffs: dict[tuple[IntegralType, int], npt.NDArray] | None = None,
) -> la.MatrixCSR:
    """Assemble bilinear form into a matrix.

    Args:
        A: The matrix to assemble into. It must have been initialized
            with the correct sparsity pattern.
        a: The bilinear form assemble.
        bcs: Boundary conditions that affect the assembled matrix.
            Degrees-of-freedom constrained by a boundary condition will
            have their rows/columns zeroed and the value ``diagonal``
            set on the diagonal.
        diag: Value to set on the matrix diagonal for Dirichlet
            boundary condition constrained degrees-of-freedom belonging
            to the same trial and test space.
        constants: Constants that appear in the form. If not provided,
            any required constants will be computed. the matrix
            diagonal.
        coeffs: Coefficients that appear in the form. If not provided,
            any required coefficients will be computed.

    Note:
        The returned matrix is not finalised, i.e. ghost values are not
        accumulated.
    """
    bcs = [] if bcs is None else [bc._cpp_object for bc in bcs]
    constants = pack_constants(a) if constants is None else constants  # type: ignore[assignment]
    coeffs = pack_coefficients(a) if coeffs is None else coeffs  # type: ignore[assignment]
    _cpp.fem.assemble_matrix(A._cpp_object, a._cpp_object, constants, coeffs, bcs)

    # If matrix is a 'diagonal'block, set diagonal entry for constrained
    # dofs
    if a.function_spaces[0] is a.function_spaces[1]:
        _cpp.fem.insert_diagonal(A._cpp_object, a.function_spaces[0], bcs, diag)
    return A


# -- Modifiers for Dirichlet conditions -----------------------------------


def apply_lifting(
    b: npt.NDArray,
    a: Sequence[Form],
    bcs: Sequence[Sequence[DirichletBC]],
    x0: Sequence[npt.NDArray] | None = None,
    alpha: float = 1,
    constants: npt.NDArray | None = None,
    coeffs: dict[tuple[IntegralType, int], npt.NDArray] | None = None,
) -> None:
    """Modify right-hand side for lifting of Dirichlet conditions.

    Consider the discrete algebraic system:

    .. math::

       \\begin{bmatrix} A_{0} & A_{1} \\end{bmatrix}
       \\begin{bmatrix}u_{0} \\\\ u_{1}\\end{bmatrix}
       = b,

    where :math:`A_{i}` is a matrix. Partitioning each vector
    :math:`u_{i}` into 'unknown' (:math:`u_{i}^{(0)}`) and prescribed
    (:math:`u_{i}^{(1)}`) groups,

    .. math::

        \\begin{bmatrix}
            A_{0}^{(0)} & A_{0}^{(1)} & A_{1}^{(0)} & A_{1}^{(1)}
        \\end{bmatrix}
        \\begin{bmatrix}
            u_{0}^{(0)} \\\\ u_{0}^{(1)} \\\\ u_{1}^{(0)} \\\\ u_{1}^{(1)}
        \\end{bmatrix}
        = b.

    If :math:`u_{i}^{(1)} = \\alpha(g_{i} - x_{i})`, where :math:`g_{i}`
    is the Dirichlet boundary condition value, :math:`x_{i}` is provided
    and :math:`\\alpha` is a constant, then

    .. math::

        \\begin{bmatrix}
            A_{0}^{(0)} & A_{0}^{(1)} & A_{1}^{(0)} & A_{1}^{(1)}
        \\end{bmatrix}
        \\begin{bmatrix}u_{0}^{(0)} \\\\ \\alpha(g_{0} - x_{0})
        \\\\ u_{1}^{(0)} \\\\ \\alpha(g_{1} - x_{1})\\end{bmatrix}
        = b.

    Rearranging,

    .. math::

        \\begin{bmatrix}A_{0}^{(0)} & A_{1}^{(0)}\\end{bmatrix}
        \\begin{bmatrix}u_{0}^{(0)} \\\\ u_{1}^{(0)}\\end{bmatrix}
        = b - \\alpha A_{0}^{(1)} (g_{0} - x_{0})
        - \\alpha A_{1}^{(1)} (g_{1} - x_{1}).

    The modified  :math:`b` vector is

    .. math::

        b \\leftarrow b - \\alpha A_{0}^{(1)} (g_{0} - x_{0})
        - \\alpha A_{1}^{(1)} (g_{1} - x_{1})

    More generally,

    .. math::
        b \\leftarrow b - \\alpha A_{i}^{(1)} (g_{i} - x_{i}).

    Args:
        b: The array to modify inplace.
        a: List of bilinear forms, where ``a[i]`` is the form that
            generates the matrix :math"`A_{i}`. All forms in ``a`` must
            share the same test function space. The trial function
            spaces can differ.
        bcs: Boundary conditions that provide the :math:`g_{i}` values.
            ``bcs1[i]`` is the sequence of boundary conditions on
            :math:`u_{i}`. Helper functions exist to build a
            list-of-lists of `DirichletBC` from a list of forms ``a``
            and a flat list of `DirichletBC` objects ``bcs``::

                bcs1 = fem.bcs_by_block(
                    fem.extract_function_spaces([a], 1),
                    bcs
                )

        x0: The array :math:`x_{i}` above. If ``None`` it is set to
            zero.
        alpha: Scalar used in the modification of ``b``.
        constants: Packed constant data appearing in the forms ``a``. If
            ``None``, the constant data will be packed by the function.
        coeffs: Packed coefficient data appearing in the forms ``a``. If
            ``None``, the coefficient data will be packed by the
            function.

    Note:
        Ghost contributions are not accumulated (not sent to owner).
        Caller is responsible for reverse-scatter to update the ghosts.

    Note:
        Boundary condition values are *not* set in ``b`` by this
        function. Use :func:`dolfinx.fem.DirichletBC.set` to set values
        in ``b``.
    """  # noqa: D301
    x0 = [] if x0 is None else x0
    constants = (
        [pack_constants(form) if form is not None else np.array([], dtype=b.dtype) for form in a]  # type: ignore[assignment]
        if constants is None
        else constants
    )
    coeffs = (
        [{} if form is None else pack_coefficients(form) for form in a]  # type: ignore[assignment]
        if coeffs is None
        else coeffs
    )
    _a = [None if form is None else form._cpp_object for form in a]
    _bcs = [[bc._cpp_object for bc in bcs0] for bcs0 in bcs]
    _cpp.fem.apply_lifting(b, _a, constants, coeffs, _bcs, x0, alpha)


def set_bc(
    b: npt.NDArray,
    bcs: Sequence[DirichletBC],
    x0: npt.NDArray | None = None,
    alpha: float = 1,
) -> None:
    """Insert boundary condition values into vector.

    Note:
        This function is deprecated.

    Only local (owned) entries are set, hence communication after
    calling this function is not required unless ghost entries need to
    be updated to the boundary condition value.
    """
    warnings.warn(
        "dolfinx.fem.assembler.set_bc is deprecated. Use dolfinx.fem.DirichletBC.set instead.",
        DeprecationWarning,
    )
    for bc in bcs:
        bc.set(b, x0, alpha)



# Copyright (C) 2017 Chris N. Richardson and Garth N. Wells
#
# This file is part of DOLFINx (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""Main module for DOLFINx"""

# flake8: noqa

import sys
# Template placeholder for injecting Windows dll directories in CI
# WINDOWSDLL

import numpy as _np

default_scalar_type: type[_np.floating | _np.complexfloating]
default_real_type: type[_np.floating]

try:
    from petsc4py import PETSc as _PETSc

    # Additional sanity check that DOLFINx was built with petsc4py support.
    import dolfinx.common

    assert dolfinx.common.has_petsc4py

    default_scalar_type = _PETSc.ScalarType  # type: ignore
    default_real_type = _PETSc.RealType  # type: ignore
except ImportError:
    default_scalar_type = _np.float64
    default_real_type = _np.float64

del _np

from dolfinx import common
from dolfinx import cpp as _cpp
from dolfinx import fem, geometry, graph, io, jit, la, log, mesh, nls, plot, typing

from dolfinx.common import (
    git_commit_hash,
    hardware_concurrency,
    has_adios2,
    has_complex_ufcx_kernels,
    has_debug,
    has_kahip,
    has_parmetis,
    has_petsc,
    has_petsc4py,
    has_ptscotch,
    has_slepc,
    has_superlu_dist,
    ufcx_signature,
)

from importlib.metadata import version

__version__ = version("fenics-dolfinx")

_cpp.common.init_logging(sys.argv)
del _cpp, sys


def get_include(user=False):
    import os

    d = os.path.dirname(__file__)
    if os.path.exists(os.path.join(d, "wrappers")):
        # Package is installed
        return os.path.join(d, "wrappers")
    else:
        # Package is from a source directory
        return os.path.join(os.path.dirname(d), "src")


__all__ = [
    "fem",
    "common",
    "geometry",
    "graph",
    "io",
    "jit",
    "la",
    "log",
    "mesh",
    "nls",
    "plot",
    "typing",
    "git_commit_hash",
    "hardware_concurrency",
    "has_adios2",
    "has_complex_ufcx_kernels",
    "has_debug",
    "has_kahip",
    "has_parmetis",
    "has_petsc",
    "has_petsc4py",
    "has_ptscotch",
    "has_slepc",
    "ufcx_signature",
]



# Copyright (C) 2025 Garth N. Wells, Jack S. Hale
#
# This file is part of DOLFINx (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""Functions for working with PETSc linear algebra objects.

Note:
    Due to subtle issues in the interaction between petsc4py memory
    management and the Python garbage collector, it is recommended that
    the PETSc method ``destroy()`` is called on returned PETSc objects
    once the object is no longer required. Note that ``destroy()`` is
    collective over the object's MPI communicator.
"""

import functools
import itertools
import typing
from collections.abc import Iterable, Sequence

from petsc4py import PETSc

import numpy as np
import numpy.typing as npt

import dolfinx
from dolfinx.common import IndexMap
from dolfinx.la import Vector

assert dolfinx.has_petsc4py

__all__ = ["assign", "create_vector", "create_vector_wrap"]


def _ghost_update(x: PETSc.Vec, insert_mode: PETSc.InsertMode, scatter_mode: PETSc.ScatterMode):  # type: ignore[name-defined]
    """Helper function for ghost updating PETSc vectors."""
    if x.getType() == PETSc.Vec.Type.NEST:  # type: ignore[attr-defined]
        for x_sub in x.getNestSubVecs():
            x_sub.ghostUpdate(addv=insert_mode, mode=scatter_mode)
            x_sub.destroy()
    else:
        x.ghostUpdate(addv=insert_mode, mode=scatter_mode)


def _zero_vector(x: PETSc.Vec):  # type: ignore[name-defined]
    """Helper function for zeroing out PETSc vectors."""
    if x.getType() == PETSc.Vec.Type.NEST:  # type: ignore[attr-defined]
        for x_sub in x.getNestSubVecs():
            with x_sub.localForm() as x_sub_local:
                x_sub_local.set(0.0)
            x_sub.destroy()
    else:
        with x.localForm() as x_local:
            x_local.set(0.0)


def create_vector_wrap(x: Vector) -> PETSc.Vec:  # type: ignore[name-defined]
    """Wrap a distributed DOLFINx vector as a PETSc vector.

    Args:
        x: The vector to wrap as a PETSc vector.

    Returns:
        A PETSc vector that shares data with ``x``.
    """
    index_map = x.index_map
    ghosts = index_map.ghosts.astype(PETSc.IntType)  # type: ignore[attr-defined]
    bs = x.block_size
    size = (index_map.size_local * bs, index_map.size_global * bs)
    return PETSc.Vec().createGhostWithArray(  # type: ignore[attr-defined]
        ghosts, x.array, size=size, bsize=bs, comm=index_map.comm
    )


def create_vector(
    maps: typing.Sequence[tuple[IndexMap, int]], kind: str | None = None
) -> PETSc.Vec:  # type: ignore[name-defined]
    """Create a PETSc vector from a sequence of maps and blocksizes.

    Three cases are supported:

    1. If ``maps=[(im_0, bs_0), ..., (im_n, bs_n)]`` is a sequence of
       indexmaps and blocksizes and ``kind`` is ``None``or is
       ``PETSc.Vec.Type.MPI``, a ghosted PETSc vector whith block structure
       described by ``(im_i, bs_i)`` is created.
       The created vector ``b`` is initialized such that on each MPI
       process ``b = [b_0, b_1, ..., b_n, b_0g, b_1g, ..., b_ng]``, where
       ``b_i`` are the entries associated with the 'owned' degrees-of-
       freedom for ``(im_i, bs_i)`` and ``b_ig`` are the 'unowned' (ghost)
       entries.

       If more than one tuple is supplied, the returned vector has an
       attribute ``_blocks`` that holds the local offsets into ``b`` for
       the (i) owned and (ii) ghost entries for each ``V[i]``. It can be
       accessed by ``b.getAttr("_blocks")``. The offsets can be used to get
       views into ``b`` for blocks, e.g.::

           >>> offsets0, offsets1, = b.getAttr("_blocks")
           >>> offsets0
           (0, 12, 28)
           >>> offsets1
           (28, 32, 35)
           >>> b0_owned = b.array[offsets0[0]:offsets0[1]]
           >>> b0_ghost = b.array[offsets1[0]:offsets1[1]]
           >>> b1_owned = b.array[offsets0[1]:offsets0[2]]
           >>> b1_ghost = b.array[offsets1[1]:offsets1[2]]

    2. If ``V=[(im_0, bs_0), ..., (im_n, bs_n)]`` is a sequence of function
       space and ``kind`` is ``PETSc.Vec.Type.NEST``, a PETSc nested vector
       (a 'nest' of ghosted PETSc vectors) is created.

    Args:
        maps: Sequence of tuples of ``IndexMap`` and the associated
            block size.
        kind: PETSc vector type (``VecType``) to create.

    Returns:
        A PETSc vector with the prescribed layout. The vector is not
        initialised to zero.
    """
    if len(maps) == 1:
        # Single space case
        index_map, bs = maps[0]
        ghosts = index_map.ghosts.astype(PETSc.IntType)  # type: ignore[attr-defined]
        size = (index_map.size_local * bs, index_map.size_global * bs)
        b = PETSc.Vec().createGhost(ghosts, size=size, bsize=bs, comm=index_map.comm)  # type: ignore
        if kind == PETSc.Vec.Type.MPI:  # type: ignore[attr-defined]
            _assign_block_data(maps, b)
        return b

    if kind is None or kind == PETSc.Vec.Type.MPI:  # type: ignore[attr-defined]
        b = dolfinx.cpp.fem.petsc.create_vector_block(maps)
        _assign_block_data(maps, b)
        return b
    elif kind == PETSc.Vec.Type.NEST:  # type: ignore[attr-defined]
        return dolfinx.cpp.fem.petsc.create_vector_nest(maps)
    else:
        raise NotImplementedError(
            "Vector type must be specified for blocked/nested assembly."
            f"Vector type '{kind}' not supported."
            "Did you mean 'nest' or 'mpi'?"
        )


@functools.singledispatch
def assign(
    x0: npt.NDArray[np.inexact] | Sequence[npt.NDArray[np.inexact]],
    x1: PETSc.Vec,  # type: ignore[name-defined]
):
    """Assign ``x0`` values to a PETSc vector ``x1``.

    Values in ``x0``, which is possibly a stacked collection of arrays,
    are assigned ``x1``. When ``x0`` holds a sequence of ``n``` arrays
    and ``x1`` has type ``NEST``, the assignment is::

              [x0[0]]
        x1 =  [x0[1]]
              [.....]
              [x0[n-1]]

    When ``x0`` holds a sequence of ``n`` arrays and ``x1`` **does
    not** have type ``NEST``, the assignment is::

              [x0_owned[0]]
        x1 =  [.....]
              [x0_owned[n-1]]
              [x0_ghost[0]]
              [.....]
              [x0_ghost[n-1]]

    Args:
        x0: An array or list of arrays that will be assigned to ``x1``.
        x1: Vector to assign values to.
    """
    if x1.getType() == PETSc.Vec.Type().NEST:  # type: ignore[attr-defined]
        x1_nest = x1.getNestSubVecs()
        for _x0, _x1 in zip(x0, x1_nest):
            with _x1.localForm() as x:
                x.array_w[:] = _x0
    else:
        with x1.localForm() as _x:
            if isinstance(x0, Sequence):
                start = 0
                for _x0 in x0:
                    end = start + _x0.shape[0]
                    _x.array_w[start:end] = _x0
                    start = end
            else:
                _x.array_w[:] = x0


@assign.register
def _(
    x0: PETSc.Vec,  # type: ignore[name-defined]
    x1: npt.NDArray[np.inexact] | Sequence[npt.NDArray[np.inexact]],
):
    """Assign PETSc vector ``x0`` values to (blocked) array(s) ``x1``.

    This function performs the reverse of the assignment performed by
    the version of :func:`.assign(x0: (npt.NDArray[np.inexact] |
    list[npt.NDArray[np.inexact]]), x1: PETSc.Vec)`.

    Args:
        x0: Vector that will have its values assigned to ``x1``.
        x1: An array or list of arrays to assign to.
    """
    if x0.getType() == PETSc.Vec.Type().NEST:  # type: ignore[attr-defined]
        x0_nest = x0.getNestSubVecs()
        for _x0, _x1 in zip(x0_nest, x1):
            with _x0.localForm() as x:
                _x1[:] = x.array_r[:]  # type: ignore[index]
    else:
        with x0.localForm() as _x0:
            if isinstance(x1, Sequence):
                start = 0
                for _x1 in x1:
                    end = start + _x1.shape[0]
                    _x1[:] = _x0.array_r[start:end]
                    start = end
            else:
                x1[:] = _x0.array_r[:]


def _assign_block_data(maps: Iterable[tuple[IndexMap, int]], vec: PETSc.Vec):  # type: ignore[name-defined]
    """Assign block data to a PETSc vector.

    Args:
        maps: Iterable of tuples each containing an ``IndexMap`` and the
        associated block size ``bs``.
        vec: PETSc vector to assign block data to.
    """
    # Early exit if the vector already has block data or is a nest vector
    if vec.getAttr("_blocks") is not None or vec.getType() == "nest":
        return

    maps = [(index_map, bs) for index_map, bs in maps]
    off_owned = tuple(
        itertools.accumulate(maps, lambda off, m: off + m[0].size_local * m[1], initial=0)
    )
    off_ghost = tuple(
        itertools.accumulate(
            maps, lambda off, m: off + m[0].num_ghosts * m[1], initial=off_owned[-1]
        )
    )
    vec.setAttr("_blocks", (off_owned, off_ghost))



# Copyright (C) 2017 Garth N. Wells
#
# This file is part of DOLFINx (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""Unit tests for MPI facilities."""

import sys

from mpi4py import MPI

from dolfinx.mesh import create_unit_square


def test_mpi_comm_wrapper():
    """Test MPICommWrapper <-> mpi4py.MPI.Comm conversion."""
    comm0 = MPI.COMM_WORLD
    m = create_unit_square(comm0, 4, 4)
    comm1 = m.comm
    assert isinstance(comm0, MPI.Comm)
    assert isinstance(comm1, MPI.Comm)


def test_mpi_comm_refcount():
    """Test MPICommWrapper <-> mpi4py.MPI.Comm reference counting."""
    comm0 = MPI.COMM_WORLD
    m = create_unit_square(comm0, 4, 4)
    comm1 = m.comm
    assert comm1 != comm0
    comm2 = m.comm
    assert comm2 == comm1

    del m
    assert sys.getrefcount(comm1) == 1 + int(sys.version_info < (3, 14))
    assert comm1.rank == comm0.rank



# Copyright (C) 2026 Jack S. Hale, Jørgen S. Dokken
#
# This file is part of DOLFINx (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
#
# The test test_curve_mesh was taken from io4dolfinx and relicensed to LGPLv3
# with permission of Jørgen S. Dokken.

from mpi4py import MPI

import numpy as np
import pytest

from basix import LagrangeVariant
from dolfinx.fem import assemble_scalar, coordinate_element, form, interpolate_geometry
from dolfinx.mesh import (
    CellType,
    create_rectangle,
    create_unit_square,
)
from ufl import ds, dx


def _assert_close_up_to_row_permutation(a, b, atol):
    """Assert rows of a and b are equal, up to a row permutation."""
    a = np.asarray(a)
    b = np.asarray(b)
    assert a.shape == b.shape
    used = [False] * len(b)
    for ra in a:
        for j, rb in enumerate(b):
            if not used[j] and np.allclose(ra, rb, atol=atol, rtol=0.0):
                used[j] = True
                break
        else:
            raise AssertionError(f"No match for {ra} in {b}")


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_interpolate_geometry_p1_to_p2(dtype):
    msh = create_unit_square(MPI.COMM_WORLD, 4, 4, dtype=dtype)
    cmap = coordinate_element(CellType.triangle, 2, dtype=dtype)

    new_msh = interpolate_geometry(msh, cmap)
    assert new_msh.geometry.cmaps[0].degree == 2

    # Topology should be the same cpp object.
    assert new_msh.topology._cpp_object is msh.topology._cpp_object

    x_old, x_new = msh.geometry.x, new_msh.geometry.x
    dm_old, dm_new = msh.geometry.dofmaps[0], new_msh.geometry.dofmaps[0]
    assert dm_old.shape[0] == dm_new.shape[0]
    assert dm_new.shape[1] == 6

    atol = 10 * np.finfo(dtype).eps
    for c in range(dm_new.shape[0]):
        v = x_old[dm_old[c]]
        # P2 geometry has original vertices + midpoints.
        expected = np.vstack([v, 0.5 * (v[0] + v[1]), 0.5 * (v[0] + v[2]), 0.5 * (v[1] + v[2])])
        _assert_close_up_to_row_permutation(x_new[dm_new[c]], expected, atol=atol)


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_interpolate_geometry_p1_roundtrip(dtype):
    msh = create_unit_square(MPI.COMM_WORLD, 4, 4, dtype=dtype)
    cmap = coordinate_element(CellType.triangle, 1, dtype=dtype)

    new_msh = interpolate_geometry(msh, cmap)

    assert new_msh.geometry.cmaps[0].degree == 1
    assert new_msh.topology._cpp_object is msh.topology._cpp_object

    x_old, x_new = msh.geometry.x, new_msh.geometry.x
    dm_old, dm_new = msh.geometry.dofmaps[0], new_msh.geometry.dofmaps[0]
    assert dm_old.shape == dm_new.shape

    atol = 10 * np.finfo(dtype).eps
    for c in range(dm_old.shape[0]):
        np.testing.assert_allclose(x_new[dm_new[c]], x_old[dm_old[c]], atol=atol, rtol=0.0)


def _curve_mesh_errors(N, degree, dtype, R, cell_type, lagrange_variant):
    """Return (area_error, circumference_error) for a degree-p curved disk mesh with N cells."""
    mesh = create_rectangle(
        MPI.COMM_WORLD,
        [[-1.0, -1.0], [1.0, 1.0]],
        [N, N],
        cell_type=cell_type,
        dtype=dtype,
    )
    original_area_form = form(1 * dx(domain=mesh), dtype=dtype)
    original_circ_form = form(1 * ds(domain=mesh), dtype=dtype)

    def transform(x):
        x_c = np.zeros_like(x)
        x_c[:, 0] = R * x[:, 0] * np.sqrt(1.0 - (x[:, 1] ** 2 / (2.0)))
        x_c[:, 1] = R * x[:, 1] * np.sqrt(1.0 - (x[:, 0] ** 2 / (2.0)))
        return x_c

    cmap = coordinate_element(cell_type, degree, variant=lagrange_variant, dtype=dtype)
    curved_mesh = interpolate_geometry(mesh, cmap)
    curved_mesh.geometry.x[:] = transform(curved_mesh.geometry.x)

    area_form = form(1 * dx(domain=curved_mesh), dtype=dtype)
    circ_form = form(1 * ds(domain=curved_mesh), dtype=dtype)

    area = curved_mesh.comm.allreduce(assemble_scalar(area_form), op=MPI.SUM)
    circ = curved_mesh.comm.allreduce(assemble_scalar(circ_form), op=MPI.SUM)

    # Interpolate curved mesh back to P1 and compare against the original P1 mesh
    # with the same transform applied — should match to near machine precision.
    cmap_linear = coordinate_element(cell_type, 1, dtype=dtype)
    linear_mesh = interpolate_geometry(curved_mesh, cmap_linear)
    linear_area_form = form(1 * dx(domain=linear_mesh), dtype=dtype)
    linear_circ_form = form(1 * ds(domain=linear_mesh), dtype=dtype)

    recovered_area = linear_mesh.comm.allreduce(assemble_scalar(linear_area_form), op=MPI.SUM)
    recovered_circ = linear_mesh.comm.allreduce(assemble_scalar(linear_circ_form), op=MPI.SUM)

    mesh.geometry.x[:] = transform(mesh.geometry.x)
    reference_area = mesh.comm.allreduce(assemble_scalar(original_area_form), op=MPI.SUM)
    reference_circ = mesh.comm.allreduce(assemble_scalar(original_circ_form), op=MPI.SUM)

    tol = 10 * np.finfo(dtype).eps
    assert np.isclose(recovered_area, reference_area, rtol=tol)
    assert np.isclose(recovered_circ, reference_circ, rtol=tol)

    return abs(area - np.pi * R**2), abs(circ - 2.0 * np.pi * R)


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
@pytest.mark.parametrize("degree", [1, 2, 3])
@pytest.mark.parametrize("R", [0.1])
@pytest.mark.parametrize("cell_type", [CellType.triangle, CellType.quadrilateral])
@pytest.mark.parametrize(
    "lagrange_variant", [LagrangeVariant.equispaced, LagrangeVariant.gll_isaac]
)
def test_curve_mesh(degree, dtype, R, cell_type, lagrange_variant):
    Ns = [4, 8, 16, 32]
    errors = [_curve_mesh_errors(N, degree, dtype, R, cell_type, lagrange_variant) for N in Ns]

    area_errors = np.array([e[0] for e in errors])
    circ_errors = np.array([e[1] for e in errors])

    hs = np.array([2.0 / N for N in Ns])
    area_rate = np.polyfit(np.log(hs), np.log(area_errors), 1)[0]
    circ_rate = np.polyfit(np.log(hs), np.log(circ_errors), 1)[0]

    # np.float32 converges to machine epsilon quickly, leading to noise
    if dtype is np.float64:
        expected_rate = degree + 1
        tolerance = 0.2
        assert area_rate >= expected_rate - tolerance, (
            f"Area convergence rate {area_rate:.2f} below expected {expected_rate}"
        )
        assert circ_rate >= expected_rate - tolerance, (
            f"Circumference convergence rate {circ_rate:.2f} below expected {expected_rate}"
        )



# Copyright (C) 2024 Matthew Scroggs and Garth N. Wells
#
# This file is part of DOLFINx (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later

from mpi4py import MPI

import numpy as np
import pytest

from dolfinx.fem import Function, functionspace
from dolfinx.mesh import create_unit_cube, create_unit_square


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
@pytest.mark.parametrize("degree", range(1, 5))
def test_dof_coords_2d(degree, dtype):
    mesh = create_unit_square(MPI.COMM_WORLD, 10, 10, dtype=dtype)
    V = functionspace(mesh, ("Lagrange", degree))
    u = Function(V, dtype=dtype)
    u.interpolate(lambda x: x[0])
    u.x.scatter_forward()
    x = V.tabulate_dof_coordinates()
    np.allclose(u.x.array, x[:, 0], atol=1e-7, rtol=1e-6)


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
@pytest.mark.parametrize("degree", range(1, 5))
def test_dof_coords_3d(degree, dtype):
    mesh = create_unit_cube(MPI.COMM_WORLD, 10, 10, 10, dtype=dtype)
    V = functionspace(mesh, ("Lagrange", degree))
    u = Function(V, dtype=dtype)
    u.interpolate(lambda x: x[0])
    u.x.scatter_forward()
    x = V.tabulate_dof_coordinates()

    eps = np.sqrt(np.finfo(dtype).eps)
    np.allclose(u.x.array, x[:, 0], atol=1e-7, rtol=eps)



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.13.8
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# # Electromagnetic modal analysis for a waveguide

# Copyright (C) 2022 Michele Castriotta, Igor Baratta, Jørgen S. Dokken
#
# ```{admonition} Download sources
# :class: download
# * {download}`Python script <./demo_half-loaded-waveguide.py>`
# * {download}`Jupyter notebook <./demo_half-loaded-waveguide.ipynb>`
# ```
#
# The demo shows how to:
# - Setup an eigenvalue problem for Maxwell's equations
# - Setup a half-loaded electromagnetic waveguide with perfect electric
#   conducting walls
# - Use SLEPc for solving eigenvalue problems
# - Verification of numerical eigenvalues with analytical modes
#
# ## Equations and problem definition
#
# In this demo, we are going to show how to solve the eigenvalue problem
# associated with a half-loaded rectangular waveguide with perfect
# electric conducting walls.
#
# First of all, let's import the modules we need for solving the
# problem:

# +
from pathlib import Path

from mpi4py import MPI
from petsc4py import PETSc

import numpy as np
from slepc4py import SLEPc

import ufl
from basix.ufl import element, mixed_element
from dolfinx import fem, plot
from dolfinx.fem.petsc import assemble_matrix
from dolfinx.mesh import CellType, create_rectangle, exterior_facet_indices, locate_entities

try:
    import pyvista

    have_pyvista = True
except ModuleNotFoundError:
    print("pyvista and pyvistaqt are required to visualise the solution")
    have_pyvista = False

if PETSc.IntType == np.int64 and MPI.COMM_WORLD.size > 1:
    print("This solver fails with PETSc and 64-bit integers because of memory errors in MUMPS.")
    # Note: when PETSc.IntType == np.int32, superlu_dist is used
    # rather than MUMPS and does not trigger memory failures.
    exit(0)

try:
    from dolfinx.io import VTXWriter

    has_vtx = True
except ImportError:
    print("VTXWriter not available, solution will not be saved.")
    has_vtx = False
# -

# ## Analytical solutions for the half-loaded waveguide
#
# The analytical solutions for the half-loaded waveguide with perfect
# electric conducting walls are described in Harrington's *Time-harmonic
# electromagnetic fields*. We will skip the full derivation, and we just
# mention that the problem can be decoupled into $\mathrm{TE}_x$ and
# $\mathrm{TM}_x$ modes, and the possible $k_z$ can be found by solving
# a set of transcendental equations, which is shown here below:
#
# $$
# \textrm{For TE}_x \textrm{ modes}:
# \begin{cases}
# &k_{x d}^{2}+\left(\frac{n \pi}{w}\right)^{2}+k_{z}^{2}=k_0^{2}
# \varepsilon_{d} \\
# &k_{x v}^{2}+\left(\frac{n \pi}{w}\right)^{2}+k_{z}^{2}=k_0^{2}
# \varepsilon_{v} \\
# & k_{x d} \cot k_{x d} d + k_{x v} \cot \left[k_{x v}(h-d)\right] = 0
# \end{cases}
# $$
#
# $$
# \textrm{For TM}_x \textrm{ modes}:
# \begin{cases}
# &k_{x d}^{2}+\left(\frac{n \pi}{w}\right)^{2}+k_{z}^{2}=
# k_0^{2} \varepsilon_{d} \\
# &k_{x v}^{2}+\left(\frac{n \pi}{w}\right)^{2}+k_{z}^{2}=
# k_0^{2} \varepsilon_{v} \\
# & \frac{k_{x d}}{\varepsilon_{d}} \tan k_{x d} d +
# \frac{k_{x v}}{\varepsilon_{v}} \tan \left[k_{x v}(h-d)\right] = 0
# \end{cases}
# $$
#
# with:
# - $\varepsilon_d\rightarrow$ dielectric permittivity
# - $\varepsilon_v\rightarrow$ vacuum permittivity
# - $w\rightarrow$ total width of the waveguide
# - $h\rightarrow$ total height of the waveguide
# - $d\rightarrow$ height of the dielectric fraction
# - $k_0\rightarrow$ vacuum wavevector
# - $k_{xd}\rightarrow$ $x$ component of the wavevector in the dielectric
# - $k_{xv}\rightarrow$ $x$ component of the wavevector in the vacuum
# - $\frac{n \pi}{w} = k_y\rightarrow$ $y$ component of the wavevector
#   for different $n$ harmonic numbers (we assume $n=1$ for the sake of
#   simplicity)
#
# Let's define the set of equations with the $\tan$ and $\cot$ function:


def TMx_condition(
    kx_d: complex, kx_v: complex, eps_d: complex, eps_v: complex, d: float, h: float
) -> float:
    """Transcendental equation for TMx modes."""
    return kx_d / eps_d * np.tan(kx_d * d) + kx_v / eps_v * np.tan(kx_v * (h - d))


def TEx_condition(kx_d: complex, kx_v: complex, d: float, h: float) -> float:
    """Transcendental equation for TEx modes."""
    return kx_d / np.tan(kx_d * d) + kx_v / np.tan(kx_v * (h - d))


# Then, we can define the `verify_mode` function, to check whether a
# certain $k_z$ satisfy the equations (below a certain threshold). In
# other words, we provide a certain $k_z$, together with the geometrical
# and optical parameters of the waveguide, and `verify_mode()` checks
# whether the last equations for the $\mathrm{TE}_x$ or $\mathrm{TM}_x$
# modes are close to $0$.


def verify_mode(
    kz: complex,
    w: float,
    h: float,
    d: float,
    lmbd0: float,
    eps_d: complex,
    eps_v: complex,
    threshold: float,
) -> np.bool_:
    """Verify if given kz satisfies the mode conditions."""
    k0 = 2 * np.pi / lmbd0
    ky = np.pi / w  # we assume n = 1
    kx_d_target = np.sqrt(k0**2 * eps_d - ky**2 + -(kz**2) + 0j)
    alpha = kx_d_target**2
    beta = alpha - k0**2 * (eps_d - eps_v)
    kx_v = np.sqrt(beta)
    kx_d = np.sqrt(alpha)
    f_tm = TMx_condition(kx_d, kx_v, eps_d, eps_v, d, h)
    f_te = TEx_condition(kx_d, kx_v, d, h)
    return np.isclose(f_tm, 0, atol=threshold) or np.isclose(f_te, 0, atol=threshold)


# We now define the domain. It is a rectangular domain with width $w$
# and height $h = 0.45w$, with the dielectric medium filling the
# lower-half of the domain, with a height of $d=0.5h$.

# +
w = 1
h = 0.45 * w
d = 0.5 * h
nx = 300
ny = int(0.4 * nx)

msh = create_rectangle(MPI.COMM_WORLD, np.array([[0, 0], [w, h]]), (nx, ny), CellType.quadrilateral)
msh.topology.create_connectivity(msh.topology.dim - 1, msh.topology.dim)
# -

# Now we can define the dielectric permittivity $\varepsilon_r$ over the
# domain as $\varepsilon_r = \varepsilon_v = 1$ in the vacuum, and as
# $\varepsilon_r = \varepsilon_d = 2.45$ in the dielectric:

# +
eps_v = 1
eps_d = 2.45

D = fem.functionspace(msh, ("DQ", 0))
eps = fem.Function(D)

cells_d = locate_entities(msh, msh.topology.dim, lambda x: x[1] <= d)

eps.x.array[:] = eps_v
eps.x.array[cells_d] = eps_d
# -

# In order to find the weak form of our problem, the starting point are
# Maxwell's equation and the perfect electric conductor condition on the
# waveguide wall:
#
# $$
# \begin{aligned}
# &\nabla \times \frac{1}{\mu_{r}} \nabla \times \mathbf{E}-k_{o}^{2}
# \epsilon_{r} \mathbf{E}=0 \quad &\text { in } \Omega\\
# &\hat{n}\times\mathbf{E} = 0 &\text { on } \Gamma
# \end{aligned}
# $$
#
# with $k_0$ and $\lambda_0 = 2\pi/k_0$ being the wavevector and the
# wavelength, which we consider fixed at $\lambda = h/0.2$. If we focus
# on non-magnetic material only, we can also use $\mu_r=1$.
#
# Now we can assume a known dependence on $z$:
#
# $$
# \mathbf{E}(x, y, z)=\left[\mathbf{E}_{t}(x, y)+\hat{z} E_{z}(x, y)\right]
# e^{-jk_z z}
# $$
#
# where $\mathbf{E}_t$ is the component of the electric field transverse
# to the waveguide axis, and $E_z$ is the component  of the electric
# field parallel to the waveguide axis, and $k_z$ represents our complex
# propagation constant.
#
# In order to pose the problem as an eigenvalue problem, we need to make
# the following substitution:
#
# $$
# \begin{aligned}
# & \mathbf{e}_t = k_z\mathbf{E}_t\\
# & e_z = -jE_z
# \end{aligned}
# $$
#
# The final weak form can be written as:
#
# $$
# \begin{aligned}
# F_{k_z}(\mathbf{e})=\int_{\Omega} &\left(\nabla_{t} \times
# \mathbf{e}_{t}\right) \cdot\left(\nabla_{t} \times
# \bar{\mathbf{v}}_{t}\right) -k_{o}^{2} \epsilon_{r} \mathbf{e}_{t} \cdot
# \bar{\mathbf{v}}_{t} \\
# &+k_z^{2}\left[\left(\nabla_{t} e_{z}+\mathbf{e}_{t}\right)
# \cdot\left(\nabla_{t} \bar{v}_{z}+\bar{\mathbf{v}}_{t}\right)-k_{o}^{2}
# \epsilon_{r} e_{z} \bar{v}_{z}\right] \mathrm{d} x = 0
# \end{aligned}
# $$
#
# Or, in a more compact form, as:
#
# $$
# \left[\begin{array}{cc}
# A_{t t} & 0 \\
# 0 & 0
# \end{array}\right]\left\{\begin{array}{l}
# \mathbf{e}_{t} \\
# e_{z}
# \end{array}\right\}=-k_z^{2}\left[\begin{array}{ll}
# B_{t t} & B_{t z} \\
# B_{z t} & B_{z z}
# \end{array}\right]\left\{\begin{array}{l}
# \mathbf{e}_{t} \\
# e_{z}
# \end{array}\right\}
# $$
#
# A problem of this kind is known as a generalized eigenvalue problem,
# where our eigenvalues are all the possible $ -k_z^2$. For further
# details about this problem, check Jin's *The Finite Element Method in
# Electromagnetics, third edition*.
#
# To write the weak form, we need to specify our function space. For
# $\mathbf{e}_t$, we can use RTCE elements (the equivalent of Nedelec
# elements on quadrilateral cells), while for $e_z$ field we can use
# Lagrange elements. This hybrid formulation is implemented with
# `mixed_element`:

degree = 1
RTCE = element("RTCE", msh.basix_cell(), degree, dtype=PETSc.RealType)
Q = element("Lagrange", msh.basix_cell(), degree, dtype=PETSc.RealType)
V = fem.functionspace(msh, mixed_element([RTCE, Q]))

# Now we can define our weak form:

# +
lmbd0 = h / 0.2
k0 = 2 * np.pi / lmbd0

et, ez = ufl.TrialFunctions(V)
vt, vz = ufl.TestFunctions(V)

a_tt = (ufl.inner(ufl.curl(et), ufl.curl(vt)) - (k0**2) * eps * ufl.inner(et, vt)) * ufl.dx
b_tt = ufl.inner(et, vt) * ufl.dx
b_tz = ufl.inner(et, ufl.grad(vz)) * ufl.dx
b_zt = ufl.inner(ufl.grad(ez), vt) * ufl.dx
b_zz = (ufl.inner(ufl.grad(ez), ufl.grad(vz)) - (k0**2) * eps * ufl.inner(ez, vz)) * ufl.dx

a = fem.form(a_tt)
b = fem.form(b_tt + b_tz + b_zt + b_zz)
# -

# Let's add the perfect electric conductor conditions on the waveguide
# wall:

# +
bc_facets = exterior_facet_indices(msh.topology)
bc_dofs = fem.locate_dofs_topological(V, msh.topology.dim - 1, bc_facets)
u_bc = fem.Function(V)
with u_bc.x.petsc_vec.localForm() as loc:
    loc.set(0)
bc = fem.dirichletbc(u_bc, bc_dofs)
# -

# ## Solve the problem in SLEPc

# Now we can solve the problem with SLEPc. First of all, we need to
# assemble our $A$ and $B$ matrices with PETSc in this way:

A = assemble_matrix(a, bcs=[bc])
A.assemble()
B = assemble_matrix(b, bcs=[bc])
B.assemble()

# Now, we need to create the eigenvalue problem in SLEPc. Our problem is
# a linear eigenvalue problem, that in SLEPc is solved with the `EPS`
# module. We can initialize this solver in the following way:

eps = SLEPc.EPS().create(msh.comm)

# We can pass to `EPS` our matrices by using the `setOperators` routine:

eps.setOperators(A, B)

# If the matrices in the problem have known properties (e.g.
# hermiticity) we can use this information in SLEPc to accelerate the
# calculation with the `setProblemType` function. For this problem,
# there is no property that can be exploited, and therefore we define it
# as a generalized non-Hermitian eigenvalue problem with the
# `SLEPc.EPS.ProblemType.GNHEP` object:

eps.setProblemType(SLEPc.EPS.ProblemType.GNHEP)

# Next, we need to specify a tolerance for the iterative solver, so that
# it knows when to stop:

tol = 1e-9
eps.setTolerances(tol=tol)

# Now we need to set the eigensolver for our problem. SLEPc offers
# different built-in algorithms, and also wrappers to external
# libraries. Some of these can only solve Hermitian problems and/or
# problems with eigenvalues in a certain portion of the spectrum.
# However, the choice of the particular method to choose to solve a
# problem is a technical discussion that is out of the scope of this
# demo, and that is more comprehensively discussed in the [SLEPc
# documentation](https://slepc.upv.es/documentation/slepc.pdf). For our
# problem, we will use the Krylov-Schur method, which we can set by
# calling the `setType` function:

eps.setType(SLEPc.EPS.Type.KRYLOVSCHUR)

# In order to accelerate the calculation of our solutions, we can also
# use a so-called *spectral transformation*, a technique which maps the
# original eigenvalues into another position of the spectrum without
# affecting the eigenvectors. In our case, we can use the
# shift-and-invert transformation with the `SLEPc.ST.Type.SINVERT`
# object:

# +
# Get ST context from eps
st = eps.getST()

# Set shift-and-invert transformation
st.setType(SLEPc.ST.Type.SINVERT)
# -

# The spectral transformation needs a target value for the eigenvalues
# we are looking for. Since the eigenvalues for our problem can be
# complex numbers, we need to specify whether we are searching for
# specific values in the real part, in the imaginary part, or in the
# magnitude. In our case, we are interested in propagating modes, and
# therefore in real $k_z$. For this reason, we can specify with the
# `setWhichEigenpairs` function that our target value will refer to the
# real part of the eigenvalue, with the `SLEPc.EPS.Which.TARGET_REAL`
# object:

eps.setWhichEigenpairs(SLEPc.EPS.Which.TARGET_REAL)

# For specifying the target value, we can use the `setTarget` function.
# Even though we cannot know a good target value a priori, we can guess
# that $k_z$ will be quite close to $k_0$ in value, for instance $k_z =
# 0.5k_0^2$. Therefore, we can set a target value of $-(0.5k_0^2)$:

eps.setTarget(-((0.5 * k0) ** 2))

# Then, we need to define the number of eigenvalues we want to
# calculate. We can do this with the `setDimensions` function, where we
# specify that we are looking for just one eigenvalue:

eps.setDimensions(nev=1)

# We can finally solve the problem with the `solve` function. To gain a
# deeper insight over the simulation, we also print an output message
# from SLEPc by calling the `view` and `errorView` function:

eps.solve()
eps.view()
eps.errorView()

# ## Verification of computed eigenvalues
# Now we can get the eigenvalues and eigenvectors calculated by SLEPc
# with the following code. We also verify if the numerical $k_z$ are
# consistent with the analytical equations of the half-loaded waveguide
# modes, with the `verify_mode()` function defined in
# `analytical_modes.py`:

# +
# Save the kz
vals = [(i, np.sqrt(-eps.getEigenvalue(i))) for i in range(eps.getConverged())]

# Sort kz by real part
vals.sort(key=lambda x: x[1].real)

eh = fem.Function(V)

kz_list = []

out_folder = Path("out_half_loaded_waveguide")
out_folder.mkdir(parents=True, exist_ok=True)

for i, kz in vals:
    # Save eigenvector in eh
    eps.getEigenpair(i, eh.x.petsc_vec)

    # Compute error for i-th eigenvalue
    error = eps.computeError(i, SLEPc.EPS.ErrorType.RELATIVE)

    # Verify, save and visualize solution
    if error < tol and np.isclose(kz.imag, 0, atol=tol):
        kz_list.append(kz)

        # Verify if kz is consistent with the analytical equations
        assert verify_mode(kz, w, h, d, lmbd0, eps_d, eps_v, threshold=1e-4)

        print(f"eigenvalue: {-(kz**2)}")
        print(f"kz: {kz}")
        print(f"kz/k0: {kz / k0}")

        eh.x.scatter_forward()

        eth, ezh = eh.split()
        eth = eh.sub(0).collapse()
        ez = eh.sub(1).collapse()

        # Transform eth, ezh into Et and Ez
        eth.x.array[:] = eth.x.array[:] / kz
        ezh.x.array[:] = ezh.x.array[:] * 1j

        gdim = msh.geometry.dim
        V_dg = fem.functionspace(msh, ("DQ", degree, (gdim,)))
        Et_dg = fem.Function(V_dg)
        Et_dg.interpolate(eth)

        if has_vtx:
            # Save solutions
            with VTXWriter(msh.comm, out_folder / f"Et_{i}.bp", Et_dg) as f:
                f.write(0.0)

            with VTXWriter(msh.comm, out_folder / f"Ez_{i}.bp", ezh) as f:
                f.write(0.0)

        # Visualize solutions with Pyvista
        if have_pyvista:
            V_cells, V_types, V_x = plot.vtk_mesh(V_dg)
            V_grid = pyvista.UnstructuredGrid(V_cells, V_types, V_x)
            Et_values = np.zeros((V_x.shape[0], 3), dtype=np.float64)
            Et_values[:, : msh.topology.dim] = Et_dg.x.array.reshape(
                V_x.shape[0], msh.topology.dim
            ).real

            V_grid.point_data["u"] = Et_values

            plotter = pyvista.Plotter()
            plotter.add_mesh(V_grid.copy(), show_edges=False)
            plotter.view_xy()
            plotter.link_views()
            if pyvista.OFF_SCREEN:
                plotter.screenshot(out_folder / "Et.png", window_size=[400, 400])
            else:
                plotter.show()

        if have_pyvista:
            V_lagr, lagr_dofs = V.sub(1).collapse()
            V_cells, V_types, V_x = plot.vtk_mesh(V_lagr)
            V_grid = pyvista.UnstructuredGrid(V_cells, V_types, V_x)
            V_grid.point_data["u"] = ezh.x.array.real[lagr_dofs[0]]
            plotter = pyvista.Plotter()
            plotter.add_mesh(V_grid.copy(), show_edges=False)
            plotter.view_xy()
            plotter.link_views()
            if pyvista.OFF_SCREEN:
                plotter.screenshot(out_folder / "Ez.png", window_size=[400, 400])
            else:
                plotter.show()



# Copyright (C) 2026 Jack S. Hale, Chris N. Richardson
#
# This file is part of DOLFINx (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""SuperLU_DIST linear solver support.

This module provides support for parallel solution of linear systems
assembled into :class:`dolfinx.la.MatrixCSR` using SuperLU_DIST.

Note:
  Users with advanced linear solver requirements should use PETSc/petsc4py.
"""

from typing import Generic, TypeVar

import numpy as np
import numpy.typing as npt

import dolfinx
import dolfinx.cpp as _cpp

if not dolfinx.has_superlu_dist:
    raise RuntimeError("DOLFINx was not built with SuperLU_DIST support.")

__all__ = ["SuperLUDistMatrix", "SuperLUDistSolver", "superlu_dist_matrix", "superlu_dist_solver"]

# As of 2026, SuperLU_DIST only supports these types, so the general Scalar
# type including np.complex64 cannot be used.
_T = TypeVar("_T", np.float32, np.float64, np.complex128)


class SuperLUDistMatrix(Generic[_T]):
    """SuperLU_DIST matrix."""

    _cpp_object: (
        _cpp.la.SuperLUDistMatrix_float32
        | _cpp.la.SuperLUDistMatrix_float64
        | _cpp.la.SuperLUDistMatrix_complex128
    )

    def __init__(self, matrix):
        """Create a SuperLU_DIST matrix.

        Args:
            matrix: C++ SuperLUDistMatrix object.

        Note:
            This initialiser is intended for internal library use only.
            User code should call :func:`superlu_dist_matrix` to create a
            :class:`SuperLUDistMatrix` object.
        """
        self._cpp_object = matrix

    @property
    def dtype(self) -> npt.DTypeLike:
        """Dtype of matrix values."""
        return self._cpp_object.dtype


def superlu_dist_matrix(A: dolfinx.la.MatrixCSR[_T]) -> SuperLUDistMatrix[_T]:
    """Create a SuperLU_DIST matrix.

    Deep copies all required data from ``A``.

    Args:
        A: Assembled matrix.

    Returns:
        A SuperLU_DIST matrix.
    """
    dtype = A.data.dtype
    if np.issubdtype(dtype, np.float32):
        stype = _cpp.la.SuperLUDistMatrix_float32
    elif np.issubdtype(dtype, np.float64):
        stype = _cpp.la.SuperLUDistMatrix_float64
    elif np.issubdtype(dtype, np.complex128):
        stype = _cpp.la.SuperLUDistMatrix_complex128
    else:
        raise NotImplementedError(f"Type {dtype} not supported.")
    return SuperLUDistMatrix(stype(A._cpp_object))


class SuperLUDistSolver(Generic[_T]):
    """SuperLU_DIST solver."""

    _cpp_object: (
        _cpp.la.SuperLUDistSolver_float32
        | _cpp.la.SuperLUDistSolver_float64
        | _cpp.la.SuperLUDistSolver_complex128
    )

    def __init__(self, solver):
        """Create a SuperLU_DIST solver.

        Args:
            solver: C++ SuperLUDistSolver object.

        Note:
            This initialiser is intended for internal library use only.
            User code should call :func:`superlu_dist_solver` to create a
            :class:`SuperLUDistSolver` object.
        """
        self._cpp_object = solver

    def set_option(self, name: str, value: str):
        """Set SuperLU_DIST option for solve.

        See SuperLU_DIST User's Guide for option names and values.

        Example::

            solver.set_option("SymmetricMode", "YES")
            solver.set_option("Trans", "NOTRANS")

        Args:
            name: Option name.
            value: Option value.
        """
        self._cpp_object.set_option(name, value)

    def set_A(self, A: SuperLUDistMatrix[_T]):
        """Set assembled left-hand side matrix.

        For advanced use with SuperLU_DIST option `Factor` allowing use of
        previously computed permutations when solving with new matrix A.

        Args:
            A: Assembled left-hand side matrix :math:`A`.
        """
        self._cpp_object.set_A(A._cpp_object)

    def solve(self, b: dolfinx.la.Vector[_T], u: dolfinx.la.Vector[_T]) -> int:
        """Solve linear system :math:`Au = b`.

        Note:
            Vectors must have size and parallel layout compatible with
            ``A``.

        Note:
            The caller must check the return integer for success
            ``(== 0)``.

        Note:
            The caller must ``u.scatter_forward()`` after the solve.

        Note:
            The values of ``A`` are modified in-place during the solve.

        Note:
            To solve with successive right-hand sides `b` the caller must
            ``solver.set_option("Factor", "FACTORED")`` after the first
            solve.

        Args:
            b: Right-hand side vector :math:`b`.
            u: Solution vector :math:`u`, overwritten during solve.

        Returns:
           SuperLU_DIST return integer from ``p*gssvx`` routine.
        """
        return self._cpp_object.solve(b._cpp_object, u._cpp_object)


def superlu_dist_solver(A: SuperLUDistMatrix[_T]) -> SuperLUDistSolver[_T]:
    """Create a SuperLU_DIST linear solver.

    Solve linear system :math:`Au = b` via LU decomposition.

    The SuperLU_DIST solver has options set to upstream defaults, except
    PrintStat (verbose solver output) set to NO.

    Args:
        A: Assembled left-hand side matrix :math:`A`.

    Returns:
        A SuperLU_DIST solver.
    """
    dtype = A.dtype
    if np.issubdtype(dtype, np.float32):
        stype = _cpp.la.SuperLUDistSolver_float32
    elif np.issubdtype(dtype, np.float64):
        stype = _cpp.la.SuperLUDistSolver_float64
    elif np.issubdtype(dtype, np.complex128):
        stype = _cpp.la.SuperLUDistSolver_complex128
    else:
        raise NotImplementedError(f"Type {dtype} not supported.")
    return SuperLUDistSolver(stype(A._cpp_object))



# Copyright (C) 2026 Jack S. Hale
#
# This file is part of DOLFINx (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""High-level problem classes using native linear algebra objects.

Users with advanced requirements should use
:mod:`dolfinx.fem.petsc`.
"""

import typing
from collections.abc import Sequence

import numpy.typing as npt

import ufl
from dolfinx import default_scalar_type
from dolfinx.fem import (
    DirichletBC,
    Form,
    Function,
    apply_lifting,
    assemble_matrix,
    assemble_vector,
    create_matrix,
    create_vector,
    form,
)
from dolfinx.la import InsertMode, MatrixCSR, Vector
from dolfinx.la.superlu_dist import superlu_dist_matrix, superlu_dist_solver
from dolfinx.mesh import EntityMap as EntityMap

__all__ = ["LinearProblem"]


class LinearProblem:
    r"""High-level class for solving linear problems using SuperLU_DIST.

    Solves problems of the form :math:`a(u, v) = L(v) \; \forall v \in V`
    using :class:`dolfinx.la.superlu_dist.SuperLUDistSolver` as the
    linear solver.

    Note:
        DOLFINx must be built with SuperLU_DIST to use this class.
    """

    def __init__(
        self,
        a: ufl.Form,
        L: ufl.Form,
        bcs: Sequence[DirichletBC] | None = None,
        u: Function | None = None,
        dtype: npt.DTypeLike = default_scalar_type,
        superlu_dist_options: dict | None = None,
        form_compiler_options: dict | None = None,
        jit_options: dict | None = None,
        entity_maps: Sequence[EntityMap] | None = None,
    ) -> None:
        """Initialize SuperLU_DIST solver for a linear variational problem.

        Args:
            a: Bilinear UFL form, the left-hand side of the variational
               problem.
            L: Linear UFL form, the right-hand side of the variational
               problem.
            bcs: Dirichlet boundary conditions to apply to the variational
               problem.
            u: Solution function. Created if not provided.
            dtype: Scalar type for form compilation. Must match
               ``u.dtype`` if ``u`` is provided.
            superlu_dist_options: Options passed to the SuperLU_DIST
                solver. See the SuperLU_DIST User's Guide for
                available options and values.
            form_compiler_options: Options used in FFCx compilation of
                all forms. Run ``ffcx --help`` at the command line to see
                all available options.
            jit_options: Options used in CFFI JIT compilation of C code
                generated by FFCx. See :func:`dolfinx.jit.ffcx_jit` for
                all available options. Takes priority over all other
                option values.
            entity_maps: If any trial functions, test functions, or
                coefficients in the form are not defined over the same mesh
                as the integration domain, a corresponding
                :class:`EntityMap <dolfinx.mesh.EntityMap>` must be
                provided.

        Example::

            problem = LinearProblem(a, L, bcs=bc,
                superlu_dist_options={"SymmetricMode": "YES"})
            u_h = problem.solve()
        """
        if u is not None:
            _dtype = u.dtype
            if dtype is not u.dtype:
                raise ValueError(f"dtype ({dtype}) does not match u.dtype ({u.dtype}).")
        else:
            _dtype = dtype

        self._a = typing.cast(
            Form,
            form(
                a,
                _dtype,
                form_compiler_options=form_compiler_options,
                jit_options=jit_options,
                entity_maps=entity_maps,
            ),
        )
        self._L = typing.cast(
            Form,
            form(
                L,
                _dtype,
                form_compiler_options=form_compiler_options,
                jit_options=jit_options,
                entity_maps=entity_maps,
            ),
        )
        self._A = create_matrix(self._a)
        self._x = create_vector(L.arguments()[0].ufl_function_space(), dtype=_dtype)
        self._b = create_vector(L.arguments()[0].ufl_function_space(), dtype=_dtype)

        self._u: Function
        if u is None:
            self._u = Function(L.arguments()[0].ufl_function_space(), dtype=_dtype)
        else:
            self._u = u

        self.bcs = [] if bcs is None else bcs

        self._superlu_dist_options = superlu_dist_options

    def solve(self) -> Function:
        """Solve the problem.

        This method updates the solution function ``u`` stored in the
        problem instance.

        Returns:
            The solution function.
        """
        # Assemble lhs
        self.A.set_value(self.A.data.dtype.type(0.0))
        assemble_matrix(self.A, self.a, bcs=self.bcs)  # type: ignore[arg-type, misc]
        self.A.scatter_reverse()

        # SuperLU_DIST solves in-place, so a deep copy of A is required.
        A_superlu_dist = superlu_dist_matrix(self.A)
        solver = superlu_dist_solver(A_superlu_dist)
        if self._superlu_dist_options is not None:
            for option, value in self._superlu_dist_options.items():
                solver.set_option(option, value)

        # Assemble rhs
        self.b.array[:] = 0.0
        assemble_vector(self.b.array, self.L)  # type: ignore[arg-type]

        # Apply boundary conditions to the rhs
        if self.bcs:
            apply_lifting(self.b.array, [self.a], bcs=[self.bcs])
            self.b.scatter_reverse(InsertMode.add)
            for bc in self.bcs:
                bc.set(self.b.array)
        else:
            self.b.scatter_reverse(InsertMode.add)

        # Solve linear system and update ghost values in the solution
        error = solver.solve(self.b, self.x)
        if error > 0:
            raise RuntimeError(f"SuperLU_DIST returned non-zero error code: {error}")
        self.x.scatter_forward()
        self.u.x.array[:] = self.x.array
        return self.u

    @property
    def L(self) -> Form:
        """The compiled linear form representing the right-hand side."""
        return self._L

    @property
    def a(self) -> Form:
        """The compiled bilinear form representing the left-hand side."""
        return self._a

    @property
    def A(self) -> MatrixCSR:
        """Left-hand side matrix."""
        return self._A

    @property
    def b(self) -> Vector:
        """Right-hand side vector."""
        return self._b

    @property
    def x(self) -> Vector:
        """Solution vector.

        Note:
            The vector does not share memory with the solution
            function ``u``.
        """
        return self._x

    @property
    def u(self) -> Function:
        """Solution function.

        Note:
            The function does not share memory with the solution
            vector ``x``.
        """
        return self._u



# Copyright (C) 2026 Jack S. Hale, Chris Richardson
#
# This file is part of DOLFINx (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""Unit tests for SuperLU_DIST."""

from mpi4py import MPI

import numpy as np
import pytest

import dolfinx
from dolfinx.fem import (
    Function,
    apply_lifting,
    assemble_matrix,
    assemble_scalar,
    assemble_vector,
    dirichletbc,
    form,
    functionspace,
    locate_dofs_topological,
)
from dolfinx.la import InsertMode
from dolfinx.mesh import create_unit_square, exterior_facet_indices
from ufl import SpatialCoordinate, TestFunction, TrialFunction, div, dx, grad, inner


@pytest.mark.parametrize("dtype", [np.float32, np.float64, np.complex128])
@pytest.mark.skipif(not dolfinx.has_superlu_dist, reason="No SuperLU_DIST")
def test_superlu_solver(dtype):
    """Manufactured Poisson and screened problem with exact solution u = x[1]**3.

    Test includes various checks that SuperLU_DIST Factor option works correctly
    for efficient re-use of permutations and factors.
    """
    from dolfinx.la.superlu_dist import superlu_dist_matrix, superlu_dist_solver

    mesh_dtype = dtype().real.dtype
    mesh = create_unit_square(MPI.COMM_WORLD, 5, 5, dtype=mesh_dtype)
    V = functionspace(mesh, ("Lagrange", 4))
    u, v = TrialFunction(V), TestFunction(V)

    a = inner(grad(u), grad(v)) * dx
    a = form(a, dtype=dtype)

    # Exact solution
    def u_ex(x):
        return x[1] ** 3

    x = SpatialCoordinate(mesh)
    f = -div(grad(u_ex(x)))
    L = inner(f, v) * dx
    L = form(L, dtype=dtype)

    u_bc = Function(V, dtype=dtype)
    u_bc.interpolate(u_ex)

    # Create Dirichlet boundary condition
    facetdim = mesh.topology.dim - 1
    mesh.topology.create_connectivity(facetdim, mesh.topology.dim)
    bndry_facets = exterior_facet_indices(mesh.topology)
    bdofs = locate_dofs_topological(V, facetdim, bndry_facets)
    bc = dirichletbc(u_bc, bdofs)

    b = assemble_vector(L)
    apply_lifting(b.array, [a], bcs=[[bc]])
    b.scatter_reverse(InsertMode.add)
    bc.set(b.array)

    a = form(a, dtype=dtype)
    A = assemble_matrix(a, bcs=[bc])
    A.scatter_reverse()

    def check_error(u_ex, uh):
        M = (u_ex(x) - uh) ** 2 * dx
        M = form(M, dtype=dtype)
        error = mesh.comm.allreduce(assemble_scalar(M), op=MPI.SUM)
        eps = np.sqrt(np.finfo(dtype).eps)
        assert np.isclose(error, 0.0, atol=eps)

    def solve_and_check(solver, b):
        uh = Function(V, dtype=dtype)
        error_code = solver.solve(b, uh.x)
        assert error_code == 0
        uh.x.scatter_forward()
        return uh

    # Standard solve
    A_superlu = superlu_dist_matrix(A)
    solver_1 = superlu_dist_solver(A_superlu)
    solver_1.set_option("SymmetricMode", "YES")
    uh_1 = solve_and_check(solver_1, b)
    check_error(u_ex, uh_1)

    # Check second solve re-using permutations from previous solve
    solver_1.set_option("Fact", "FACTORED")
    solve_and_check(solver_1, b)
    uh_2 = solve_and_check(solver_1, b)
    check_error(u_ex, uh_2)

    # Check can do same solve with MatrixCSR A again.
    A_superlu_2 = superlu_dist_matrix(A)
    solver_2 = superlu_dist_solver(A_superlu_2)
    solver_2.set_option("SymmetricMode", "YES")
    uh_3 = solve_and_check(solver_2, b)
    check_error(u_ex, uh_3)

    # Assemble a new operator and re-use row permutation from solve with old
    # operator.
    gamma = 1e-1
    a_2 = (inner(grad(u), grad(v)) + gamma * inner(u, v)) * dx
    a_2 = form(a_2, dtype=dtype)

    f_2 = -div(grad(u_ex(x))) + gamma * u_ex(x)
    L_2 = inner(f_2, v) * dx
    L_2 = form(L_2, dtype=dtype)

    b_2 = assemble_vector(L_2)
    apply_lifting(b_2.array, [a_2], bcs=[[bc]])
    b_2.scatter_reverse(InsertMode.add)
    bc.set(b_2.array)

    A_3 = assemble_matrix(a_2, bcs=[bc])
    A_3.scatter_reverse()

    A_superlu_3 = superlu_dist_matrix(A_3)
    solver_2.set_A(A_superlu_3)
    solver_2.set_option("Fact", "SamePattern")
    uh_2 = solve_and_check(solver_2, b_2)
    check_error(u_ex, uh_2)



# Copyright (C) 2021 Chris Richardson and Igor Baratta
#
# This file is part of DOLFINx (https://www.fenicsproject.org)
#
# SPDX-License-Identifier:    LGPL-3.0-or-later
"""Unit tests for the KrylovSolver interface."""

from mpi4py import MPI

import numpy as np
import pytest

from basix.ufl import element
from dolfinx import default_real_type, la
from dolfinx.fem import Function, functionspace
from dolfinx.mesh import create_unit_square


@pytest.mark.parametrize(
    "e",
    [
        element("Lagrange", "triangle", 1, dtype=default_real_type),
        element("Lagrange", "triangle", 1, shape=(2,), dtype=default_real_type),
    ],
)
def test_scatter_forward(e):
    mesh = create_unit_square(MPI.COMM_WORLD, 5, 5)
    V = functionspace(mesh, e)
    u = Function(V)
    bs = V.dofmap.bs

    u.interpolate(lambda x: [x[i] for i in range(bs)])

    # Forward scatter should have no effect
    w0 = u.x.array.copy()
    u.x.scatter_forward()
    assert np.allclose(w0, u.x.array)

    # Fill local array with the mpi rank
    u.x.array.fill(MPI.COMM_WORLD.rank)
    w0 = u.x.array.copy()
    u.x.scatter_forward()

    # Now the ghosts should have the value of the rank of the owning
    # process
    ghost_owners = u.function_space.dofmap.index_map.owners
    ghost_owners = np.repeat(ghost_owners, bs)
    local_size = u.function_space.dofmap.index_map.size_local * bs
    assert np.allclose(u.x.array[local_size:], ghost_owners)


@pytest.mark.parametrize(
    "e",
    [
        element("Lagrange", "triangle", 1, dtype=default_real_type),
        element("Lagrange", "triangle", 1, shape=(2,), dtype=default_real_type),
    ],
)
def test_scatter_reverse(e):
    comm = MPI.COMM_WORLD
    mesh = create_unit_square(MPI.COMM_WORLD, 5, 5)
    V = functionspace(mesh, e)
    u = Function(V)
    bs = V.dofmap.bs

    u.interpolate(lambda x: [x[i] for i in range(bs)])

    # Reverse scatter (insert) should have no effect
    w0 = u.x.array.copy()
    u.x.scatter_reverse(la.InsertMode.insert)
    assert np.allclose(w0, u.x.array)

    # Fill with MPI rank, and sum all entries in the vector (including
    # ghosts)
    u.x.array.fill(comm.rank)
    all_count0 = MPI.COMM_WORLD.allreduce(u.x.array.sum(), op=MPI.SUM)

    # Reverse scatter (add)
    u.x.scatter_reverse(la.InsertMode.add)
    num_ghosts = V.dofmap.index_map.num_ghosts
    ghost_count = MPI.COMM_WORLD.allreduce(num_ghosts * comm.rank, op=MPI.SUM)

    # New count should have gone up by the number of ghosts times their
    # rank on all processes
    all_count1 = MPI.COMM_WORLD.allreduce(u.x.array.sum(), op=MPI.SUM)
    assert all_count1 == (all_count0 + bs * ghost_count)


@pytest.mark.parametrize(
    "dtype",
    [
        np.float32,
        np.float64,
        np.complex64,
        np.complex128,
        np.int8,
        np.int32,
        np.int64,
    ],
)
def test_vector_from_index_map_scatter_forward(dtype):
    comm = MPI.COMM_WORLD
    mesh = create_unit_square(comm, 5, 5)

    for d in range(mesh.topology.dim):
        mesh.topology.create_entities(d)
        im = mesh.topology.index_map(d)
        vector = la.vector(im, dtype=dtype)
        vector.array[: im.size_local] = np.arange(*im.local_range, dtype=dtype)
        vector.scatter_forward()
        global_idxs = im.local_to_global(np.arange(im.size_local + im.num_ghosts, dtype=np.int32))
        global_idxs = np.asarray(global_idxs, dtype)

        assert np.all(vector.array == global_idxs)
